/* ptrace.c */
/* By Ross Biro 1/23/92 */
/*
 * Pentium III FXSR, SSE support
 *	Gareth Hughes <gareth@valinux.com>, May 2000
 */

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/smp.h>
#include <linux/smp_lock.h>
#include <linux/errno.h>
#include <linux/ptrace.h>
#include <linux/user.h>

#include <asm/uaccess.h>
#include <asm/pgtable.h>
#include <asm/system.h>
#include <asm/processor.h>
#include <asm/i387.h>
#include <asm/debugreg.h>

#ifdef CONFIG_MOSIX
#include <linux/hpc.h>
#endif /* CONFIG_MOSIX */

/*
 * does not yet catch signals sent when the child dies.
 * in exit.c or in signal.c.
 */

/* determines which flags the user has access to. */
/* 1 = access 0 = no access */
#define FLAG_MASK 0x00044dd5

/* set's the trap flag. */
#define TRAP_FLAG 0x100

/*
 * Offset of eflags on child stack..
 */
#define EFL_OFFSET ((EFL-2)*4-sizeof(struct pt_regs))

/*
 * this routine will get a word off of the processes privileged stack. 
 * the offset is how far from the base addr as stored in the TSS.  
 * this routine assumes that all the privileged stacks are in our
 * data space.
 */   
#ifdef CONFIG_MOSIX
inline int get_stack_long(struct task_struct *task, int offset)
#else
static inline int get_stack_long(struct task_struct *task, int offset)
#endif /* CONFIG_MOSIX */
{
	unsigned char *stack;

#ifdef CONFIG_MOSIX
	if(task != current)
		lock_mosix();
	if(task->mosix.dflags & DDEPUTY)
	{
		if(task != current)
			unlock_mosix();
		return(request_process(task, NULL, PR_PTRACE_GET_STACK_LONG,
			offset));
	}
	if(task == current)
		mosix_obtain_registers(ALL_REGISTERS);
	else
		task_lock(task);
	if(!task->thread.saved_esp0)	/* unless in VM86 mode */
		stack = (char *)(mos_to_regs(&task->mosix) + 1);
	else
#endif /* CONFIG_MOSIX */
	stack = (unsigned char *)task->thread.esp0;
	stack += offset;
#ifdef CONFIG_MOSIX
	{
		int res = *((int *)stack);
		if(task != current)
		{
			task_unlock(task);
			unlock_mosix();
		}
		return(res);
	}
#else
	return (*((int *)stack));
#endif /* CONFIG_MOSIX */
}

/*
 * this routine will put a word on the processes privileged stack. 
 * the offset is how far from the base addr as stored in the TSS.  
 * this routine assumes that all the privileged stacks are in our
 * data space.
 */
#ifdef CONFIG_MOSIX
inline int put_stack_long(struct task_struct *task, int offset,
#else
static inline int put_stack_long(struct task_struct *task, int offset,
#endif /* CONFIG_MOSIX */
	unsigned long data)
{
	unsigned char * stack;

#ifdef CONFIG_MOSIX
	if(task != current)
		lock_mosix();
	if(task->mosix.dflags & DDEPUTY)
	{
		if(task != current)
			unlock_mosix();
		return(request_process_arg2(task, NULL,
			PR_PTRACE_PUT_STACK_LONG, offset, data));
	}
	if(task == current)
		mosix_obtain_registers(ALL_REGISTERS);
	else
		task_lock(task);
	if(!task->thread.saved_esp0)	/* unless in VM86 mode */
		stack = (char *)(mos_to_regs(&task->mosix) + 1);
	else
#endif /* CONFIG_MOSIX */
	stack = (unsigned char *) task->thread.esp0;
	stack += offset;
	*(unsigned long *) stack = data;
#ifdef CONFIG_MOSIX
	if(task != current)
	{
		task_unlock(task);
		unlock_mosix();
	}
#endif /* CONFIG_MOSIX */
	return 0;
}

static int putreg(struct task_struct *child,
	unsigned long regno, unsigned long value)
{
	switch (regno >> 2) {
		case FS:
			if (value && (value & 3) != 3)
				return -EIO;
			child->thread.fs = value;
			return 0;
		case GS:
			if (value && (value & 3) != 3)
				return -EIO;
			child->thread.gs = value;
			return 0;
		case DS:
		case ES:
			if (value && (value & 3) != 3)
				return -EIO;
			value &= 0xffff;
			break;
		case SS:
		case CS:
			if ((value & 3) != 3)
				return -EIO;
			value &= 0xffff;
			break;
		case EFL:
			value &= FLAG_MASK;
			value |= get_stack_long(child, EFL_OFFSET) & ~FLAG_MASK;
			break;
	}
	if (regno > GS*4)
		regno -= 2*4;
	put_stack_long(child, regno - sizeof(struct pt_regs), value);
	return 0;
}

#ifdef CONFIG_MOSIX
void
ptrace_putregs(unsigned long *data)
{
	register int i;

	for (i = 0; i < FRAME_SIZE ; i++)
		putreg(current, i << 2, data[i]);
}
#endif /* CONFIG_MOSIX */

static unsigned long getreg(struct task_struct *child,
	unsigned long regno)
{
	unsigned long retval = ~0UL;

	switch (regno >> 2) {
		case FS:
			retval = child->thread.fs;
			break;
		case GS:
			retval = child->thread.gs;
			break;
		case DS:
		case ES:
		case SS:
		case CS:
			retval = 0xffff;
			/* fall through */
		default:
			if (regno > GS*4)
				regno -= 2*4;
			regno = regno - sizeof(struct pt_regs);
			retval &= get_stack_long(child, regno);
	}
	return retval;
}

/*
 * Called by kernel/ptrace.c when detaching..
 *
 * Make sure the single step bit is not set.
 */
void ptrace_disable(struct task_struct *child)
{ 
	long tmp;

	tmp = get_stack_long(child, EFL_OFFSET) & ~TRAP_FLAG;
	put_stack_long(child, EFL_OFFSET, tmp);
}

#ifdef CONFIG_MOSIX
void
ptrace_getregs(unsigned long *data)
{
	register int i;

	for (i = 0; i < FRAME_SIZE ; i++)
		data[i] = getreg(current, i << 2);
}
#endif /* CONFIG_MOSIX */

asmlinkage int sys_ptrace(long request, long pid, long addr, long data)
{
	struct task_struct *child;
	struct user * dummy = NULL;
	int i, ret;

	lock_kernel();
	ret = -EPERM;
	if (request == PTRACE_TRACEME) {
		/* are we already being traced? */
		if (current->ptrace & PT_PTRACED)
			goto out;
		/* set the ptrace bit in the process flags. */
		current->ptrace |= PT_PTRACED;
#ifdef CONFIG_MOSIX_DFSA
		tell_process(current, DREQ_NOTUPTODATE);
#endif /* CONFIG_MOSIX_DFSA */
		ret = 0;
		goto out;
	}
	ret = -ESRCH;
	read_lock(&tasklist_lock);
	child = find_task_by_pid(pid);
	if (child)
		get_task_struct(child);
	read_unlock(&tasklist_lock);
	if (!child)
		goto out;

	ret = -EPERM;
	if (pid == 1)		/* you may not mess with init */
		goto out_tsk;

	if (request == PTRACE_ATTACH) {
		ret = ptrace_attach(child);
		goto out_tsk;
	}

	ret = ptrace_check_attach(child, request == PTRACE_KILL);
	if (ret < 0)
		goto out_tsk;

	switch (request) {
	/* when I and D space are separate, these will need to be fixed. */
	case PTRACE_PEEKTEXT: /* read word at location addr. */ 
	case PTRACE_PEEKDATA: {
		unsigned long tmp;
		int copied;

		copied = access_process_vm(child, addr, &tmp, sizeof(tmp), 0);
		ret = -EIO;
		if (copied != sizeof(tmp))
			break;
		ret = put_user(tmp,(unsigned long *) data);
		break;
	}

	/* read the word at location addr in the USER area. */
	case PTRACE_PEEKUSR: {
		unsigned long tmp;

		ret = -EIO;
		if ((addr & 3) || addr < 0 || 
		    addr > sizeof(struct user) - 3)
			break;

#ifdef CONFIG_MOSIX
		lock_mosix();
		if(child->mosix.dflags & DDEPUTY)
		{
			unlock_mosix();
			if(request_process(child, &tmp, PR_PTRACE_PEEKUSER, addr))
				goto out;
		}
		else
		{
#endif /* CONFIG_MOSIX */
		tmp = 0;  /* Default return condition */
		if(addr < FRAME_SIZE*sizeof(long))
			tmp = getreg(child, addr);
		if(addr >= (long) &dummy->u_debugreg[0] &&
		   addr <= (long) &dummy->u_debugreg[7]){
			addr -= (long) &dummy->u_debugreg[0];
			addr = addr >> 2;
			tmp = child->thread.debugreg[addr];
		}
#ifdef CONFIG_MOSIX
		unlock_mosix();
		}
#endif /* CONFIG_MOSIX */
		ret = put_user(tmp,(unsigned long *) data);
		break;
	}

	/* when I and D space are separate, this will have to be fixed. */
	case PTRACE_POKETEXT: /* write the word at location addr. */
	case PTRACE_POKEDATA:
		ret = 0;
		if (access_process_vm(child, addr, &data, sizeof(data), 1) == sizeof(data))
			break;
		ret = -EIO;
		break;

	case PTRACE_POKEUSR: /* write the word at location addr in the USER area */
		ret = -EIO;
		if ((addr & 3) || addr < 0 || 
		    addr > sizeof(struct user) - 3)
			break;

		if (addr < FRAME_SIZE*sizeof(long)) {
			ret = putreg(child, addr, data);
			break;
		}
		/* We need to be very careful here.  We implicitly
		   want to modify a portion of the task_struct, and we
		   have to be selective about what portions we allow someone
		   to modify. */

		  ret = -EIO;
		  if(addr >= (long) &dummy->u_debugreg[0] &&
		     addr <= (long) &dummy->u_debugreg[7]){

			  if(addr == (long) &dummy->u_debugreg[4]) break;
			  if(addr == (long) &dummy->u_debugreg[5]) break;
			  if(addr < (long) &dummy->u_debugreg[4] &&
			     ((unsigned long) data) >= TASK_SIZE-3) break;
			  
			  if(addr == (long) &dummy->u_debugreg[7]) {
				  data &= ~DR_CONTROL_RESERVED;
				  for(i=0; i<4; i++)
					  if ((0x5f54 >> ((data >> (16 + 4*i)) & 0xf)) & 1)
						  goto out_tsk;
			  }

			  addr -= (long) &dummy->u_debugreg;
			  addr = addr >> 2;
#ifdef CONFIG_MOSIX
			  lock_mosix();
			  if(child->mosix.dflags & DDEPUTY)
			  {
				unlock_mosix();
				if(request_process_arg2(child, NULL,
					PR_PTRACE_POKEUSER, addr, data))
					goto out;
			  }
			  else
			  {
#endif /* CONFIG_MOSIX */
			  child->thread.debugreg[addr] = data;
#ifdef CONFIG_MOSIX
			  unlock_mosix();
			  }
#endif /* CONFIG_MOSIX */
			  ret = 0;
		  }
		  break;

	case PTRACE_SYSCALL: /* continue and stop at next (return from) syscall */
	case PTRACE_CONT: { /* restart after signal. */
		long tmp;

		ret = -EIO;
		if ((unsigned long) data > _NSIG)
			break;
		if (request == PTRACE_SYSCALL)
			child->ptrace |= PT_TRACESYS;
		else
			child->ptrace &= ~PT_TRACESYS;
		child->exit_code = data;
#ifdef CONFIG_MOSIX
		if(child->mosix.dflags & DDEPUTY)
		{
			if(request_process(child, NULL, PR_PTRACE_CONT,
								request))
				goto out;
		}
		else
		{
#endif /* CONFIG_MOSIX */
	/* make sure the single step bit is not set. */
		tmp = get_stack_long(child, EFL_OFFSET) & ~TRAP_FLAG;
		put_stack_long(child, EFL_OFFSET,tmp);
#ifdef CONFIG_MOSIX
		}
#endif /* CONFIG_MOSIX */
		wake_up_process(child);
		ret = 0;
		break;
	}

/*
 * make the child exit.  Best I can do is send it a sigkill. 
 * perhaps it should be put in the status that it wants to 
 * exit.
 */
	case PTRACE_KILL: {
		long tmp;

		ret = 0;
#ifdef CONFIG_MOSIX
		if (LOGICAL_STATE(child) == TASK_ZOMBIE) /* already dead */
#else
		if (child->state == TASK_ZOMBIE)	/* already dead */
#endif /* CONFIG_MOSIX */
			break;
		child->exit_code = SIGKILL;
		/* make sure the single step bit is not set. */
		tmp = get_stack_long(child, EFL_OFFSET) & ~TRAP_FLAG;
		put_stack_long(child, EFL_OFFSET, tmp);
		wake_up_process(child);
		break;
	}

	case PTRACE_SINGLESTEP: {  /* set the trap flag. */
		long tmp;

		ret = -EIO;
		if ((unsigned long) data > _NSIG)
			break;
		child->ptrace &= ~PT_TRACESYS;
		if ((child->ptrace & PT_DTRACE) == 0) {
			/* Spurious delayed TF traps may occur */
			child->ptrace |= PT_DTRACE;
		}
#ifdef CONFIG_MOSIX
		if(child->mosix.dflags & DDEPUTY)
		{
			if(request_process(child, NULL, PR_PTRACE_SINGLE_STEP,
									0))
				goto out;
		}
		else
		{
#endif /* CONFIG_MOSIX */
		tmp = get_stack_long(child, EFL_OFFSET) | TRAP_FLAG;
		put_stack_long(child, EFL_OFFSET, tmp);
#ifdef CONFIG_MOSIX
		}
#endif /* CONFIG_MOSIX */
		child->exit_code = data;
		/* give it a chance to run. */
		wake_up_process(child);
		ret = 0;
		break;
	}

	case PTRACE_DETACH:
		/* detach a process that was attached. */
		ret = ptrace_detach(child, data);
		break;

	case PTRACE_GETREGS: { /* Get all gp regs from the child. */
	  	if (!access_ok(VERIFY_WRITE, (unsigned *)data, FRAME_SIZE*sizeof(long))) {
			ret = -EIO;
			break;
		}
#ifdef CONFIG_MOSIX
		/* NOTE: even when child does not look like a DEPUTY,
		 * it could easily become one while we wait for a page
		 * on __put_user ... therefore, we do this ALWAYS:
		 */
		{
			unsigned long t[FRAME_SIZE];

			if(request_process(child, &t, PR_PTRACE_GETREGS, 0))
				ret = -EIO;
			else
				copy_to_user((void *)data, t, sizeof(t));
		}
#else
		for ( i = 0; i < FRAME_SIZE*sizeof(long); i += sizeof(long) ) {
			__put_user(getreg(child, i),(unsigned long *) data);
			data += sizeof(long);
		}
#endif /* CONFIG_MOSIX */
		ret = 0;
		break;
	}

	case PTRACE_SETREGS: { /* Set all gp regs in the child. */
#ifndef CONFIG_MOSIX
		unsigned long tmp;
#endif /* CONFIG_MOSIX */
	  	if (!access_ok(VERIFY_READ, (unsigned *)data, FRAME_SIZE*sizeof(long))) {
			ret = -EIO;
			break;
		}
#ifdef CONFIG_MOSIX
		/* NOTE: even when child does not look like a DEPUTY,
		 * it could easily become one while we wait for a page
		 * on __get_user ... therefore, we do this ALWAYS:
		 */
		{
			unsigned long t[FRAME_SIZE];

			copy_from_user(t, (void *)data, sizeof(t));
			ret = request_process(child, &t, PR_PTRACE_SETREGS, 0)
								? -EIO : 0;
		}
#else
		for ( i = 0; i < FRAME_SIZE*sizeof(long); i += sizeof(long) ) {
			__get_user(tmp, (unsigned long *) data);
			putreg(child, i, tmp);
			data += sizeof(long);
		}
		ret = 0;
#endif /* CONFIG_MOSIX */
		break;
	}

	case PTRACE_GETFPREGS: { /* Get the child FPU state. */
		if (!access_ok(VERIFY_WRITE, (unsigned *)data,
			       sizeof(struct user_i387_struct))) {
			ret = -EIO;
			break;
		}
		ret = 0;
#ifdef CONFIG_MOSIX
		/* DEPUTY or NOT - the child-process may become DEPUTY
		 * while we page-fault on "__copy_to_user",
		 * so we always do it the long way:
		 */
		{
			struct user_i387_struct f;

			if(request_process(child, &f, PR_PTRACE_GETFPREGS, 0))
			{
				ret = -EIO;
				goto out;
			}
			__copy_to_user((void *)data, &f,
				sizeof(struct user_i387_struct));
		}
#else
		if ( !child->used_math )
			load_empty_fpu(child);
		get_fpregs((struct user_i387_struct *)data, child);
#endif /* CONFIG_MOSIX */
		break;
	}

	case PTRACE_SETFPREGS: { /* Set the child FPU state. */
		if (!access_ok(VERIFY_READ, (unsigned *)data,
			       sizeof(struct user_i387_struct))) {
			ret = -EIO;
			break;
		}
#ifdef CONFIG_MOSIX
		/* DEPUTY or NOT - the child-process may become DEPUTY
		 * while we page-fault on "__copy_from_user",
		 * so we always do it the long way:
		 */
		{
			struct user_i387_struct f;

			__copy_from_user(&f, (void *)data, sizeof(f));
			ret = request_process(child, &f, PR_PTRACE_SETFPREGS, 0)
								? -EIO : 0;
		}
#else
		child->used_math = 1;
		set_fpregs(child, (struct user_i387_struct *)data);
		ret = 0;
#endif /* CONFIG_MOSIX */
		break;
	}

	case PTRACE_GETFPXREGS: { /* Get the child extended FPU state. */
		if (!access_ok(VERIFY_WRITE, (unsigned *)data,
			       sizeof(struct user_fxsr_struct))) {
			ret = -EIO;
			break;
		}
#ifdef CONFIG_MOSIX
		ret = 0;
		/* DEPUTY or NOT - the child-process may become DEPUTY
		 * while we page-fault on "__copy_to_user",
		 * so we always do it the long way:
		 */
		{
			struct user_fxsr_struct f;

			if(request_process(child, &f, PR_PTRACE_GETFPXREGS, 0))
			{
				ret = -EIO;
				goto out;
			}
			__copy_to_user((void *)data, &f,
				sizeof(struct user_fxsr_struct));
		}
#else
		if ( !child->used_math )
			load_empty_fpu(child);
		ret = get_fpxregs((struct user_fxsr_struct *)data, child);
#endif /* CONFIG_MOSIX */
		break;
	}

	case PTRACE_SETFPXREGS: { /* Set the child extended FPU state. */
		if (!access_ok(VERIFY_READ, (unsigned *)data,
			       sizeof(struct user_fxsr_struct))) {
			ret = -EIO;
			break;
		}
#ifdef CONFIG_MOSIX
		/* DEPUTY or NOT - the child-process may become DEPUTY
		 * while we page-fault on "__copy_from_user",
		 * so we always do it the long way:
		 */
		{
			struct user_fxsr_struct f;

			__copy_from_user(&f, (void *)data, sizeof(f));
			ret = request_process(child, &f, PR_PTRACE_SETFPXREGS,
								0) ? -EIO : 0;
		}
#else
		child->used_math = 1;
		ret = set_fpxregs(child, (struct user_fxsr_struct *)data);
#endif /* CONFIG_MOSIX */
		break;
	}

	case PTRACE_SETOPTIONS: {
		if (data & PTRACE_O_TRACESYSGOOD)
			child->ptrace |= PT_TRACESYSGOOD;
		else
			child->ptrace &= ~PT_TRACESYSGOOD;
		ret = 0;
		break;
	}

	default:
		ret = -EIO;
		break;
	}
out_tsk:
	free_task_struct(child);
out:
	unlock_kernel();
	return ret;
}

#ifdef CONFIG_MOSIX
/*
 * The following routines are parts of sys_ptrace performed by the
 * child process on itself.
 * Should the relevant code in "sys_ptrace" change, the following routines
 * must also be modified accordingly.
 */
unsigned long
ptrace_peekuser(long addr)
{
	struct user * dummy = NULL;
	unsigned long tmp;

	tmp = 0;  /* Default return condition */
	if(addr < FRAME_SIZE*sizeof(long))
		tmp = getreg(current, addr);
	if(addr >= (long) &dummy->u_debugreg[0] &&
	   addr <= (long) &dummy->u_debugreg[7]){
		addr -= (long) &dummy->u_debugreg[0];
		addr = addr >> 2;
		tmp = current->thread.debugreg[addr];
	};
	return(tmp);
}

void
ptrace_pokeuser(long addr, long data)
{
	current->thread.debugreg[addr] = data;
#define loaddebug(tsk,register) \
		__asm__("movl %0,%%db" #register  \
			: /* no output */ \
			:"r" (tsk->thread.debugreg[register]))
	/* unlike the original ptrace code, we are doing this for ourselves,
	 * and there may be no "switch_to" before we go to user-mode again,
	 * so here we go:
	 */
	if(current->thread.debugreg[7])
	{
		loaddebug(current, 0);
		loaddebug(current, 1);
		loaddebug(current, 2);
		loaddebug(current, 3);
		loaddebug(current, 6);
		loaddebug(current, 7);
	}
}

void
ptrace_cont(int request)
{
	unsigned long tmp;
	struct task_struct *tsk = current;

	tmp = get_stack_long(tsk, EFL_OFFSET) & ~TRAP_FLAG;
	put_stack_long(tsk, EFL_OFFSET, tmp);
	if (request == PTRACE_SYSCALL)
		tsk->ptrace |= PT_TRACESYS;
	else
		tsk->ptrace &= ~PT_TRACESYS;
	if (request == PTRACE_DETACH)
		tsk->ptrace &= ~PT_PTRACED;
}

void
ptrace_single_step(void)
{
	unsigned long tmp;

	tmp = get_stack_long(current, EFL_OFFSET) | TRAP_FLAG;
	put_stack_long(current, EFL_OFFSET, tmp);
	current->ptrace |= PT_PTRACED | PT_DTRACE;
#ifdef CONFIG_MOSIX_DFSA
	tell_process(current, DREQ_NOTUPTODATE);
#endif /* CONFIG_MOSIX_DFSA */
}

void
ptrace_getfpregs(struct user_i387_struct *to)
{
	struct task_struct *p = current;

	unlazy_fpu(p);
	if (!p->used_math)
		load_empty_fpu(p);
	get_fpregs(to, p);
}

void
ptrace_getfpxregs(struct user_fxsr_struct *to)
{
	struct task_struct *p = current;

	unlazy_fpu(p);
	if (!p->used_math)
		load_empty_fpu(p);
	get_fpxregs(to, p);
}

void
ptrace_setfpregs(struct user_i387_struct *from)
{
	struct task_struct *p = current;

	clear_fpu(p);
	p->used_math = 1;
	set_fpregs(p, from);
}

void
ptrace_setfpxregs(struct user_fxsr_struct *from)
{
	struct task_struct *p = current;

	clear_fpu(p);
	p->used_math = 1;
	set_fpxregs(p, from);
}
#endif /* CONFIG_MOSIX */

#ifdef CONFIG_MOSIX
/*
 * When a process starts a system call in PT_TRACESYS mode, it must run
 * the full sequence: A=syscall_trace, B=the_system_call, C=syscall_trace.
 * Since our process can migrate at any of those stages, we maintain 2 flags
 * that tell us where to resume in the sequence.
 * In the particular case of DTRACESYS1, the trace was already complete
 * before the migration.  In the case of DTRACESYS2, the system-call was already
 * complete before the migration and the code in "entry.S" prevents it from
 * re-running.
 */
asmlinkage void
syscall_trace(void)
{
	extern asmlinkage void do_syscall_trace(void);

	switch(current->mosix.dflags & (DTRACESYS1|DTRACESYS2))
	{
		case 0:
			current->mosix.dflags |= DTRACESYS1;
			break;
		case DTRACESYS1:
			return;
		case DTRACESYS2:
		case DTRACESYS1|DTRACESYS2:
			current->mosix.dflags &= ~(DTRACESYS1|DTRACESYS2);
	}
	if(current->mosix.dflags & DREMOTE)
		mosix_remote_syscall_trace();
	else
		do_syscall_trace();
}

asmlinkage void do_syscall_trace(void)
#else
asmlinkage void syscall_trace(void)
#endif /* CONFIG_MOSIX */
{
	if ((current->ptrace & (PT_PTRACED|PT_TRACESYS)) !=
			(PT_PTRACED|PT_TRACESYS))
		return;
	/* the 0x80 provides a way for the tracing parent to distinguish
	   between a syscall stop and SIGTRAP delivery */
	current->exit_code = SIGTRAP | ((current->ptrace & PT_TRACESYSGOOD)
					? 0x80 : 0);
	current->state = TASK_STOPPED;
	notify_parent(current, SIGCHLD);
	schedule();
	/*
	 * this isn't the same as continuing with a signal, but it will do
	 * for normal use.  strace only continues with a signal if the
	 * stopping signal is not SIGTRAP.  -brl
	 */
	if (current->exit_code) {
		send_sig(current->exit_code, current, 1);
		current->exit_code = 0;
	}
}
