/*
 * Copyright (C) 2000, 2001, Amnon Barak (amnon@cs.huji.ac.il)
 * Some sections copyright 2002 by Moshe Bar
 * Permission to use this software is hereby granted under the terms of the
 * GNU General Public License, as published by the Free Software Foundation.
 *
 * THIS  SOFTWARE  IS  PROVIDED IN ITS  "AS IS"  CONDITION, WITH NO WARRANTY
 * WHATSOEVER. NO LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE WILL BE ACCEPTED.
 */
/*
 * Author(s): Amnon Shiloh, Moshe Bar
 */

#include <linux/config.h>
#ifdef CONFIG_MOSIX_FS
#include <linux/slab.h>
#include <linux/mfs.h>
#include <linux/mfs_socket.h>

spinlock_t mfs_client_lock = SPIN_LOCK_UNLOCKED;

#define MFS_CLIENT_OPTCONS	50
#define	MFS_MAX_HOGGING_A_NODE	8

struct mfs_client_contact
{
	struct mfs_client_contact *next;
	int to;
	struct socket *sock;
	int serial;
	int has_mh;
	struct mfs_message_header mh;
} *mfs_busy_clients, *mfs_connected_clients, *mfs_closing_clients;

int mfs_client_serial;

DECLARE_WAIT_QUEUE_HEAD(wait_for_busy_contact);
int waiting_for_busy_contact;

int
mfs_client_send(struct mfs_client_contact *s, int type, void *header, int hdln,
			void *dat1, int len1, void *dat2, int len2)
{
	return(mfs_send(s->sock, 0, s->serial,
				type, header, hdln, dat1, len1, dat2, len2));
}

int
mfs_client_receive(struct mfs_client_contact *s,
					void **header, void **dat1, void **dat2)
{
	if(s->has_mh)
	{
		s->has_mh = 0;
if(mfs_debug > 1)
printk("mfs_client_receive: continuing with serial %d, type %d (%d/%d/%d)\n",
s->mh.serial, s->mh.type, s->mh.hdsz, s->mh.dat1sz, s->mh.dat2sz);
		return(mfs_receive(s->sock, MFS_MUST_MATCH, &s->serial, &s->mh,
							header, dat1, dat2));
	}
	else
		return(mfs_receive(s->sock, MFS_SKIP_OTHERS, &s->serial, NULL,
							header, dat1, dat2));
}

int
mfs_client_contact_has_data(struct mfs_client_contact *con)
{
	return(mfs_socket_has_data(con->sock));
}

void
delayed_close_client(struct mfs_client_contact *con)
{
	mfs_send(con->sock, 1, 0, MFS_CLOSE, NULL, 0, NULL, 0, NULL,0);
	spin_lock(&mfs_client_lock);
	con->next = mfs_closing_clients;
	mfs_closing_clients = con;
	spin_unlock(&mfs_client_lock);
}

void
mfs_client_check_closed(struct mfs_client_contact **head)
{
	struct mfs_client_contact *c, *prev, *next;
	struct mfs_client_contact *dispose = NULL;

	spin_lock(&mfs_client_lock);
	for(prev = NULL, c = *head ; c ; c = next)
	{
		next = c->next;
		if(mfs_socket_is_closed(c->sock))
		{
			if(prev)
				prev->next = next;
			else
				*head = next;
			c->next = dispose;
			dispose = c;
		}
		else
			prev = c;
	}
	spin_unlock(&mfs_client_lock);
	for(; dispose ; dispose = next)
	{
		next = dispose->next;
		mfs_close_socket(dispose->sock);
		kfree(dispose);
	}
}

void
mfs_release_contact(struct mfs_client_contact *con)
{
	struct mfs_client_contact *s, *prev;
	unsigned int bcnt = 0, pcnt = 0, tcnt = 0;
	int pe = con->to;
	unsigned int found = 0;

	spin_lock(&mfs_client_lock);
	for(prev = NULL , s = mfs_busy_clients ; s ; )
	{
		tcnt++;
		if(s == con)
		{
			s = s->next;
			if(prev)
				prev->next = s;
			else
				mfs_busy_clients = s;
			found = 1;
		}
		else
		{
			if(s->to == pe && ++bcnt >= MFS_MAX_HOGGING_A_NODE &&
									found)
				break;
			prev = s;
			s = s->next;
		}
	}
	if(!found)
		printk("mfs_release_contact: not listed!\n");
	if((pcnt = bcnt) < MFS_MAX_HOGGING_A_NODE)
	for(s = mfs_connected_clients ; s ; s = s->next)
	{
		tcnt++;
		if(s->to == pe && ++pcnt == MFS_MAX_HOGGING_A_NODE)
			break;
	}
	if((tcnt >= MFS_CLIENT_OPTCONS && pcnt) ||
						pcnt >= MFS_MAX_HOGGING_A_NODE)
	{
		spin_unlock(&mfs_client_lock);
		delayed_close_client(con);
		if(bcnt >= MFS_MAX_HOGGING_A_NODE)
			return;
		spin_lock(&mfs_client_lock);
	}
	else
	{
		con->next = mfs_connected_clients;
		mfs_connected_clients = con;
	}
	if(waiting_for_busy_contact == pe ||
		waiting_for_busy_contact == MFS_UNKNOWN_PE)
	{
		waiting_for_busy_contact = 0;
		wake_up(&wait_for_busy_contact);
	}
	spin_unlock(&mfs_client_lock);
}

static inline int
node_is_hogged(int pe)
{
	unsigned int n = 0;
	struct mfs_client_contact *s;

	for(s = mfs_busy_clients ; s ; s = s->next)
	if(s->to == pe && ++n == MFS_MAX_HOGGING_A_NODE)
		return(1);
	return(0);
}

struct mfs_client_contact *
mfs_new_request(int pe, int type, void *header, int hdln, void *dat1, int len1,
							void *dat2, int len2)
{
	struct mfs_client_contact *s, *prev, *prevprev, *con;
	unsigned int tried_a_new = 0;
	unsigned int cnt;
	DECLARE_WAITQUEUE(wait, current);

	if(pe == MFS_BAD_PE)
		return(ERR_PTR(-ESTALE));
	spin_lock(&mfs_client_lock);
	for(prev = prevprev = NULL , cnt = 0 , s = mfs_connected_clients ; s ;
					prevprev = prev, prev = s , s = s->next)
	if(s->to == pe)
		break;
	else
		cnt++;
	if(s)
	{
		found_a_connected:
		if(prev)
			prev->next = s->next;
		else
			mfs_connected_clients = s->next;
		try_it:
		con = s;
		con->to = pe;
		con->next = mfs_busy_clients;
		mfs_busy_clients = con;
		con->serial = ++mfs_client_serial;
		spin_unlock(&mfs_client_lock);
		if(!mfs_send(con->sock, 1, con->serial, type, header, hdln,
						dat1, len1, dat2, len2) &&
			mfs_receive(con->sock, MFS_JUST_HEADER, &con->serial,
					&con->mh, NULL, NULL, NULL) >= 0)
		{
			con->has_mh = 1;
if(mfs_debug > 1)
printk("mfs_new_request: first reply will be serial %d, type %d (%d/%d/%d)\n",
con->mh.serial, con->mh.type, con->mh.hdsz, con->mh.dat1sz, con->mh.dat2sz);
			return(con);
		}
		spin_lock(&mfs_client_lock);
		for(prev = NULL , s = mfs_busy_clients ; s && s != con ;
							prev = s , s = s->next)
			;
		if(s)
		{
			if(prev)
				prev->next = s->next;
			else
				mfs_busy_clients = s->next;
			if(waiting_for_busy_contact == pe ||
				waiting_for_busy_contact == MFS_UNKNOWN_PE)
			{
				waiting_for_busy_contact = 0;
				wake_up(&wait_for_busy_contact);
			}
			spin_unlock(&mfs_client_lock);
			mfs_close_socket(s->sock);
			kfree(s);
			spin_lock(&mfs_client_lock);
		}
		else
			printk("mfs_new_request: contact disappeared!\n");
	}
	else if(cnt >= MFS_CLIENT_OPTCONS)
	{
		/* trim out the last one (also last used) */
		prevprev->next = NULL;
		spin_unlock(&mfs_client_lock);
		delayed_close_client(prev);
		mfs_client_check_closed(&mfs_closing_clients);
		spin_lock(&mfs_client_lock);
	}
	if(tried_a_new || !PE)
	{
		spin_unlock(&mfs_client_lock);
		return(ERR_PTR(-EREMOTE));
	}
	if(node_is_hogged(pe))
	{
		add_wait_queue(&wait_for_busy_contact, &wait);
		while(1)
		{
			set_current_state(TASK_UNINTERRUPTIBLE);
			if(waiting_for_busy_contact &&
				waiting_for_busy_contact != pe)
				waiting_for_busy_contact = MFS_UNKNOWN_PE;
			else
				waiting_for_busy_contact = pe;
			if(!node_is_hogged(pe))
				break;
			spin_unlock(&mfs_client_lock);
			if(!PE)
			{
				remove_wait_queue(&wait_for_busy_contact, &wait);
				set_current_state(TASK_RUNNING);
				return(NULL);
			}
			schedule_timeout(HZ);
			spin_lock(&mfs_client_lock);
			for(prev = NULL , s = mfs_connected_clients ; s ;
							prev = s , s = s->next)
			if(s->to == pe)
			{
				remove_wait_queue(&wait_for_busy_contact,&wait);
				set_current_state(TASK_RUNNING);
				goto found_a_connected;
			}
		}
		remove_wait_queue(&wait_for_busy_contact, &wait);
		set_current_state(TASK_RUNNING);
	}
	spin_unlock(&mfs_client_lock);
	tried_a_new++;
	if(!(s = (struct mfs_client_contact *)kmalloc(sizeof(*s), GFP_KERNEL)))
		return(ERR_PTR(-ENOMEM));
	s->sock = mfs_connect_to(pe);
	s->has_mh = 0;
	if(!IS_ERR(s->sock))
	{
		spin_lock(&mfs_client_lock);
		goto try_it;
	}
	con = (struct mfs_client_contact *)s->sock;
	kfree(s);
	return(con);
}

void
mfs_monitor_client_contacts(void)
{
	mfs_client_check_closed(&mfs_connected_clients);
	mfs_client_check_closed(&mfs_closing_clients);
}
#endif /* CONFIG_MOSIX_FS */
