/*
 * Copyright (C) 2000, 2001, Amnon Barak (amnon@cs.huji.ac.il)
 * Some sections copyright 2002 by Moshe Bar
 * Permission to use this software is hereby granted under the terms of the
 * GNU General Public License, as published by the Free Software Foundation.
 *
 * Author(s): Amnon Shiloh, Moshe Bar
 */

#include <linux/sched.h>
#ifdef CONFIG_MOSIX_FS
#include <asm/uaccess.h>
#include <linux/slab.h>
#include <linux/mfs.h>
#include <linux/utime.h>
#include <linux/smp_lock.h>
#include <linux/quotaops.h>

extern struct dentry *lookup_create(struct nameidata *, int);
extern int chown_common(struct dentry *, uid_t, gid_t);

#define	MAX_READLINK	10240

void
mfs_prepare_other(struct mfs_completion *c, struct nameidata *ond)
{
	struct dentry *other = ond->dentry;
	struct mfs_dinfo *info;

	if(other->d_sb != mfs_super ||
		!(info = (struct mfs_dinfo *)other->d_fsdata))
		c->other_pe = 0;
	else
	{
		c->other_pe = info->pe;
		c->other = info->handle;
		if(atomic_read(&other->d_count) > 1)
			c->dispose_other = 0;
		else
		{
			kfree(info);
			other->d_fsdata = NULL;
			c->dispose_other = 1;
		}
	}
}

void
mfs_prepare_utime(struct mfs_completion *c, time_t *pa, time_t *pm)
{
	if(get_user(c->args.times.a, pa) || get_user(c->args.times.m, pm))
	{
		c->needed = 0;
		c->error = -EFAULT;
	}
}

void
mfs_prepare_completion(struct mfs_completion *c, struct nameidata *nd)
{
	struct dentry *base = nd->dentry;
	struct utimbuf *ut;
	struct timeval *uts;
	struct mfs_dinfo *info;

	c->done = 0;
	c->other_pe = 0;
	c->dispose_other = 0;
	c->more_data = NULL;
	c->more_data_len = 0;
	c->error = 0;
	c->ro = mfs_is_ro;
	c->dispose_base = 0;
	if((info = (struct mfs_dinfo *)base->d_fsdata))
	{
		c->base_pe = info->pe;
		c->base = info->handle;
		if(atomic_read(&base->d_count) == 1)
		{
			kfree(info);
			base->d_fsdata = NULL;
			c->dispose_base = 1;
		}
	}
	else
	{
		c->base_pe = 0;
		c->base = MFS_ROOT_INO;
	}
	if(!(c->needed = ((nd->flags & LOOKUP_COMPLETE) != 0)))
		return;
	c->args = nd->complete_args;
	switch(c->func = nd->express_function)
	{
		case EF_OPEN:
		case EF_ACCESS:
		case EF_UNLINK:
		case EF_MKDIR:
		case EF_RMDIR:
		case EF_TRUNCATE:
		case EF_CHOWN:
		case EF_CHMOD:
		case EF_MKNOD:
		case EF_STATFS:
			return;
		case EF_RENAME:
			if(nd->complete_args.oldnd->last_type == LAST_NORM)
			{
				c->more_data = (char *)
					nd->complete_args.oldnd->last.name;
				c->more_data_len = strlen(c->more_data)+1;
			}
			/* fall through */
		case EF_LINK:
			mfs_prepare_other(c, nd->complete_args.oldnd);
			return;
		case EF_SYMLINK:
			c->more_data = nd->complete_parg;
			c->more_data_len = strlen(c->more_data)+1;
			return;
		case EF_READLINK:
			if(c->args.buffer.bufsiz > MAX_READLINK)
				c->args.buffer.bufsiz = MAX_READLINK;
			return;
		case EF_UTIME:
			ut = (struct utimbuf *)nd->complete_parg;
			if(ut)
				mfs_prepare_utime(c, &ut->actime, &ut->modtime);
			else
				c->other_pe = 1;
			return;
		case EF_UTIMES:
			uts = (struct timeval *)nd->complete_parg;
			if(uts)
				mfs_prepare_utime(c, &uts[0].tv_sec,
								&uts[1].tv_sec);
			else
				c->other_pe = 1;
			return;
		case EF_STAT:
			nd->complete_parg = NULL;
			return;
		default:
			c->needed = 0;
			return;
	}
}

int
mfs_low_complete_link(struct mfs_completion *c, struct nameidata *nd)
{
	struct nameidata ond;
	int err;
	struct dentry *new_dentry;

	if(c->other_pe != MFS_PE)
		return(-EXDEV);
	if((err = mfs_to_local(c->other, &ond)))
		return(err);
	err = -EXDEV;
	if(ond.mnt != nd->mnt)
		goto out;
	new_dentry = lookup_create(nd, 0);
	if(IS_ERR(new_dentry))
		err = PTR_ERR(new_dentry);
	else
	{
		err = vfs_link(ond.dentry, nd->dentry->d_inode, new_dentry);
		dput(new_dentry);
	}
	up(&nd->dentry->d_inode->i_sem);
	out:
	mfs_lput(ond.dentry, ond.mnt);
	return(err);
}

int
mfs_low_complete_unlink(struct mfs_completion *c, struct nameidata *nd)
{
	int err;
	struct dentry *dentry;

	if(nd->last_type != LAST_NORM)
		return(-EISDIR);
	down(&nd->dentry->d_inode->i_sem);
	dentry = lookup_hash(&nd->last, nd->dentry);
	if(IS_ERR(dentry))
		err = PTR_ERR(dentry);
	else
	{
		if(!dentry->d_inode)
			err = -ENOENT;
		else if(nd->last.name[nd->last.len])
			err = S_ISDIR(dentry->d_inode->i_mode) ?
							-EISDIR : -ENOTDIR;
		else
			err = vfs_unlink(nd->dentry->d_inode, dentry);
		dput(dentry);
	}
	up(&nd->dentry->d_inode->i_sem);
	return(err);
}

int
mfs_low_complete_symlink(struct mfs_completion *c, struct nameidata *nd)
{
	int err;
	struct dentry *dentry;

	dentry = lookup_create(nd, 0);
	if(IS_ERR(dentry))
		err = PTR_ERR(dentry);
	else
	{
		err = vfs_symlink(nd->dentry->d_inode, dentry, c->more_data);
		dput(dentry);
	}
	up(&nd->dentry->d_inode->i_sem);
	return(err);
}

int
mfs_low_complete_mkdir(struct mfs_completion *c, struct nameidata *nd)
{
	int err;
	struct dentry *dentry;

	dentry = lookup_create(nd, 1);
	if(IS_ERR(dentry))
		err = PTR_ERR(dentry);
	else
	{
		err = vfs_mkdir(nd->dentry->d_inode, dentry, c->args.ints.i1);
		dput(dentry);
	}
	up(&nd->dentry->d_inode->i_sem);
	return(err);
}

int
mfs_low_complete_mknod(struct mfs_completion *c, struct nameidata *nd)
{
	int err;
	struct dentry *dentry;
	int mode = c->args.ints.i1;

	dentry = lookup_create(nd, 1);
	if(IS_ERR(dentry))
		err = PTR_ERR(dentry);
	else
	{
		switch(mode & S_IFMT)
		{
			case 0:
			case S_IFREG:
				err = vfs_create(nd->dentry->d_inode, dentry,
									mode);
				break;
			case S_IFCHR:
			case S_IFBLK:
			case S_IFIFO:
			case S_IFSOCK:
			case S_IFDIR:
				err = -EPERM;
				break;
			default:
				err = -EINVAL;
				break;
		}
		dput(dentry);
	}
	up(&nd->dentry->d_inode->i_sem);
	return(err);
}

int
mfs_low_complete_rmdir(struct mfs_completion *c, struct nameidata *nd)
{
	int err;
	struct dentry *dentry;

	switch(nd->last_type)
	{
		case LAST_NORM:
			break;
		case LAST_DOTDOT:
			return(-ENOTEMPTY);
		default:
			return(-EBUSY);
	}
	down(&nd->dentry->d_inode->i_sem);
	dentry = lookup_hash(&nd->last, nd->dentry);
	if(IS_ERR(dentry))
		err = PTR_ERR(dentry);
	else
	{
		err = vfs_rmdir(nd->dentry->d_inode, dentry);
		dput(dentry);
	}
	up(&nd->dentry->d_inode->i_sem);
	return(err);
}

int
mfs_low_complete_rename(struct mfs_completion *c, struct nameidata *newnd)
{
	struct nameidata ond;
	int err;
	struct dentry *old_dir, *new_dir;
	struct dentry *old_dentry, *new_dentry;

	if(c->other_pe != MFS_PE)
		return(-EXDEV);
	if((err = mfs_to_local(c->other, &ond)))
		return(err);
	if(c->more_data_len)
	{
		unsigned int hash = init_name_hash();
		const unsigned char *n;

		ond.last_type = LAST_NORM;
		ond.last.name = c->more_data;
		ond.last.len = strlen(c->more_data);
		for(n = ond.last.name ; *n && *n != '/' ; n++)
			hash = partial_name_hash(*n, hash);
		ond.last.hash = end_name_hash(hash);
	}
	err = -EXDEV;
	if(ond.mnt != newnd->mnt)
		goto out;
	err = -EBUSY;
	if (ond.last_type != LAST_NORM || newnd->last_type != LAST_NORM)
		goto out;
	old_dir = ond.dentry;
	new_dir = newnd->dentry;
	double_lock(new_dir, old_dir);

	old_dentry = lookup_hash(&ond.last, old_dir);
	err = PTR_ERR(old_dentry);
	if (IS_ERR(old_dentry))
		goto out1;
	/* source must exist */
	err = -ENOENT;
	if (!old_dentry->d_inode)
		goto out2;
	/* unless the source is a directory trailing slashes give -ENOTDIR */
	if (!S_ISDIR(old_dentry->d_inode->i_mode)) {
		err = -ENOTDIR;
		if (ond.last.name[ond.last.len])
			goto out2;
		if (newnd->last.name[newnd->last.len])
			goto out2;
	}
	new_dentry = lookup_hash(&newnd->last, new_dir);
	err = PTR_ERR(new_dentry);
	if (IS_ERR(new_dentry))
		goto out2;

	lock_kernel();
	err = vfs_rename(old_dir->d_inode, old_dentry, new_dir->d_inode,
								new_dentry);
	unlock_kernel();

	dput(new_dentry);
	out2:
	dput(old_dentry);
	out1:
	double_up(&new_dir->d_inode->i_sem, &old_dir->d_inode->i_sem);
	out:
	mfs_lput(ond.dentry, ond.mnt);
	return(err);
}

int
mfs_low_complete_readlink(struct mfs_completion *c, struct nameidata *nd)
{
	char *buf;
	mm_segment_t old_fs = get_fs();
	int err;
	struct inode *ip = nd->dentry->d_inode;
	int sz = c->args.buffer.bufsiz;

	if(!ip)
		return(-ENOENT);
	if(!ip->i_op || !ip->i_op->readlink)
		return(-EINVAL);
	if(ip->i_op && ip->i_op->revalidate &&
		(err = ip->i_op->revalidate(nd->dentry)))
		return(err);
	if(!(buf = kmalloc(sz, GFP_KERNEL)))
		return(-ENOMEM);
	UPDATE_ATIME(ip);
	set_fs(KERNEL_DS);
	if(!ip->i_op || !ip->i_op->readlink)
		err = -EINVAL;
	else
		err = ip->i_op->readlink(nd->dentry, buf, sz);
	set_fs(old_fs);
	if(err > 0)
	{
		c->more_data = buf;
		c->more_data_len = err;
	}
	else
		kfree(buf);
	return(err);
}

int
mfs_low_complete_truncate(struct mfs_completion *c, struct nameidata *nd)
{
	int err;
	struct inode *ip = nd->dentry->d_inode;

	if(!S_ISREG(ip->i_mode))
		return(-EACCES);
	if((err = permission(ip, MAY_WRITE)))
		return(err);
	if(IS_RDONLY(ip))
		return(-EROFS);
	if(IS_IMMUTABLE(ip) || IS_APPEND(ip))
		return(-EPERM);
	if((err = get_write_access(ip)))
		return(err);
	if(!(err = locks_verify_truncate(ip, NULL, c->args.len)))
	{
		DQUOT_INIT(ip);
		err = do_truncate(nd->dentry, c->args.len);
	}
	put_write_access(ip);
	return(err);
}

int
mfs_low_complete_utime(struct mfs_completion *c, struct nameidata *nd)
{
	int err;
	struct iattr newattrs;
	struct inode *ip = nd->dentry->d_inode;

	if(IS_RDONLY(ip))
		return(-EROFS);
	newattrs.ia_valid = ATTR_CTIME | ATTR_MTIME | ATTR_ATIME;
	if(c->other_pe)
	{
		if((err = permission(ip, MAY_WRITE)))
			if(current->fsuid != ip->i_uid && (err = permission(ip, MAY_WRITE)))
			return(err);
	}
	else
	{
		newattrs.ia_atime = c->args.times.a;
		newattrs.ia_mtime = c->args.times.m;
		newattrs.ia_valid |= ATTR_ATIME_SET | ATTR_MTIME_SET;
	}
	return(notify_change(nd->dentry, &newattrs));
}

int
mfs_low_complete_chmod(struct mfs_completion *c, struct nameidata *nd)
{
	struct iattr newattrs;
	mode_t mode = c->args.mode;
	struct inode *ip = nd->dentry->d_inode;

	if(IS_RDONLY(ip))
		return(-EROFS);
	if(IS_IMMUTABLE(ip) || IS_APPEND(ip))
		return(-EPERM);
	if (mode == (mode_t) -1)
		mode = ip->i_mode;
	newattrs.ia_mode = (mode & S_IALLUGO) | (ip->i_mode & ~S_IALLUGO);
	newattrs.ia_valid = ATTR_MODE | ATTR_CTIME;
	return(notify_change(nd->dentry, &newattrs));
}

int
mfs_low_complete_stat(struct mfs_completion *c, struct nameidata *nd)
{
	int err;
	struct inode *ip = nd->dentry->d_inode;

	if (ip->i_op && ip->i_op->revalidate &&
		(err = ip->i_op->revalidate(nd->dentry)))
		return(err);
	if(!(c->more_data = kmalloc(sizeof(struct subinode), GFP_KERNEL)))
		return(-ENOMEM);
	c->more_data_len = sizeof(struct subinode);
	ip = nd->dentry->d_inode;
	mfs_ip_to_subip((struct subinode *)c->more_data, ip);
	((struct subinode *)c->more_data)->si_ino = global_inum(ip);
	return(0);
}

mfs_handle_t
mfs_low_complete(struct mfs_completion *c, struct nameidata *nd)
{
	mfs_handle_t result;

	if(!c->needed)
		goto export;
	c->error = 0;
	c->done = 1;
	if(c->ro)
		current->mosix.dirty_bits |= MFSARG_RONLY;
	switch(c->func)
	{
		case EF_OPEN:
			c->error = open_namei(NULL, c->args.ints.i1,
							c->args.ints.i2, nd);
			break;
		case EF_ACCESS:
			c->error = permission(nd->dentry->d_inode,
							c->args.ints.i1);
			break;
		case EF_LINK:
			c->error = mfs_low_complete_link(c, nd);
			break;
		case EF_UNLINK:
			c->error = mfs_low_complete_unlink(c, nd);
			break;
		case EF_SYMLINK:
			c->error = mfs_low_complete_symlink(c, nd);
			c->more_data = NULL;
			c->more_data_len = 0;
			break;
		case EF_MKDIR:
			c->error = mfs_low_complete_mkdir(c, nd);
			break;
		case EF_MKNOD:
			c->error = mfs_low_complete_mknod(c, nd);
			break;
		case EF_RMDIR:
			c->error = mfs_low_complete_rmdir(c, nd);
			break;
		case EF_RENAME:
			c->error = mfs_low_complete_rename(c, nd);
			c->more_data = NULL;
			c->more_data_len = 0;
			break;
		case EF_READLINK:
			c->error = mfs_low_complete_readlink(c, nd);
			break;
		case EF_TRUNCATE:
			c->error = mfs_low_complete_truncate(c, nd);
			break;
		case EF_UTIME:
		case EF_UTIMES:
			c->error = mfs_low_complete_utime(c, nd);
			break;
		case EF_CHMOD:
			c->error = mfs_low_complete_chmod(c, nd);
			break;
		case EF_CHOWN:
			c->error = chown_common(nd->dentry, c->args.ids.uid,
							c->args.ids.gid);
			break;
		case EF_STAT:
			c->error = mfs_low_complete_stat(c, nd);
			break;
		case EF_STATFS:
			c->error = mfs_low_complete_statfs(c, nd);
			break;
		default:
			c->done = 0;
			break;
	}
	current->mosix.dirty_bits &= ~MFSARG_RONLY;
	if(c->dispose_other && c->other_pe == MFS_PE)
	{
		c->dispose_other = 0;
		mfs_low_dispose_handle(c->other);
	}
	if(c->dispose_base && c->base_pe == MFS_PE)
	{
		c->dispose_base = 0;
		mfs_low_dispose_handle(c->base);
	}	
	if(c->done && (c->func != EF_OPEN || c->error))
		return(0);
	export:
	if(!(result = local_to_mfs(nd)))
	{
		if(c->more_data)
		{
			kfree(c->more_data);
			c->more_data = NULL;
			c->more_data_len = 0;
		}
		c->error = -ENOMEM;
		path_release(nd);
		if(c->needed)
			c->done = 1;
	}
	return(result);
}

int mfs_low_complete_statfs(struct mfs_completion *c, struct nameidata *nd)
{

	int err;
	struct inode *ip = nd->dentry->d_inode;

	if(!ip)
		return (-ENOENT);
	if(!(c->more_data = kmalloc(sizeof(struct statfs), GFP_KERNEL)))
		 return(-ENOMEM);
	err = vfs_statfs(ip->i_sb, (struct statfs *)c->more_data);
	if (err)
	{
		kfree(c->more_data);
		c->more_data = NULL;
	}
	else
		c->more_data_len = sizeof(struct statfs);
	return(err);
}


void
mfs_low_complete_error(struct mfs_completion *c, int err)
{
	if(c->dispose_base && c->base_pe == MFS_PE)
	{
		c->dispose_base = 0;
		mfs_low_dispose_handle(c->base);
	}
	if(c->dispose_other && c->other_pe == MFS_PE)
	{
		c->dispose_other = 0;
		mfs_low_dispose_handle(c->other);
	}
	c->error = err;
	c->more_data = NULL;	/* in case of EF_SYMLINK */
	c->more_data_len = 0;
	if(c->needed)
		c->done = 1;
}

int
mfs_check_completion(struct mfs_completion *c, struct nameidata *nd)
{
	struct inode *inode;
	struct subinode *subip;

	if(c->dispose_other)
		mfs_dispose_handle(c->other_pe, c->other);
	if(c->dispose_base)
		mfs_dispose_handle(c->base_pe, c->base);
	if(!c->done)
		return(0);
	nd->complete_flags |= COMPLETE_DONE;
	if(c->error && (c->func != EF_READLINK || c->error < 0))
		return(1);
	switch(c->func)
	{
		case EF_OPEN:
			return(0);
		case EF_STAT:
			nd->complete_parg = NULL;
			if(!(subip = (struct subinode *)c->more_data))
			{
				if(!c->error)
					c->error = -ENOMEM;
				return(1);
			}
			if((inode = (struct inode *)
				    kmalloc(sizeof(struct inode), GFP_KERNEL)))
			{
				path_release(nd);
				mfs_subip_to_ip(inode, subip);
				inode->i_ino =
				    ((struct subinode *)c->more_data)->si_ino;
				inode->i_dev = mfs_super->s_dev;
				inode->u.mfs_i.pe = c->other_pe;
				nd->complete_parg = inode;
				/* the caller must later kfree this! */
			}
			else if(!c->error)
				c->error = -ENOMEM;
			kfree(subip);
			return(1);
		case EF_READLINK:
			if(c->more_data)
			{
				if(c->error >= 0 &&
					copy_to_user((void *)c->args.buffer.buf,
					c->more_data, c->more_data_len))
					c->error = -EFAULT;
				kfree(c->more_data);
				path_release(nd);
			}
			else if(c->error >= 0)
				c->error = -ENOMEM;
			return(1);
		case EF_STATFS:
			if(c->more_data) {
				if(copy_to_user((void *)c->args.statfs,
						c->more_data, sizeof(struct statfs)))
					c->error = -EFAULT;
				kfree(c->more_data);
				path_release(nd);
				
			}
			
		default:
			path_release(nd);
		case 0:
			return(1);
	}
}

int
mfs_link_complete(struct mfs_completion *c, int pe, struct nameidata *nd)
{
	int result;
	char node[6];
	int len, copy;
	struct inode *ip;

	if(c->dispose_other)
		mfs_dispose_handle(c->other_pe, c->other);
	if(c->dispose_base)
		mfs_dispose_handle(c->base_pe, c->base);
	if(!c->needed)
		return(-EPERM);
	nd->complete_flags |= COMPLETE_DONE;
	switch(c->func)
	{
		case EF_READLINK:
			sprintf(node, "%d", pe);
			len = strlen(node);
			result = copy = min(len, c->args.buffer.bufsiz);
			if(copy_to_user((void *)c->args.buffer.buf, node, copy))
				return(-EFAULT);
			break;
		case EF_STAT:
			if(!(ip = (struct inode *)
				    kmalloc(sizeof(*ip), GFP_KERNEL)))
				return(-ENOMEM);
			nd->complete_parg = ip;
			memset(ip, 0, sizeof(*ip));
			ip->i_dev = mfs_super->s_dev;
			ip->i_ino = pe;
			ip->u.mfs_i.pe = pe;
			ip->i_mode = S_IFLNK | 0555;
			ip->i_nlink = 1;
			for(len = pe ; len ; len /= 10)
				ip->i_size++;
			ip->i_version = pe;
			result = 0;
			break;
		default:
			return(-EPERM);
	}
	path_release(nd);
	return(result);
}
#endif /* CONFIG_MOSIX_FS */
