/* all changes since Feb 12, 2002 copyright by Moshe Bar (moshe@moshebar.com)
 * Copyright (C) 2000, 2001, Amnon Barak (amnon@cs.huji.ac.il)
 * Some sections copyright 2002 by Moshe Bar
 * Permission to use this software is hereby granted under the terms of the
 * GNU General Public License, as published by the Free Software Foundation.
 *
 * THIS  SOFTWARE  IS  PROVIDED IN ITS  "AS IS"  CONDITION, WITH NO WARRANTY
 * WHATSOEVER. NO LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE WILL BE ACCEPTED.
 */
/*
 * Author(s): Amnon Shiloh, Moshe Bar
 */

#include <linux/sched.h>
#ifdef CONFIG_MOSIX_FS
#include <linux/hpc.h>
#include <linux/slab.h>
#include <linux/mfs.h>
#include <asm/uaccess.h>

static struct semaphore mfsc_sem = __MUTEX_INITIALIZER(mfsc_sem);

mfs_handle_t mfs_unique;
struct dentry *mfs_root;
struct vfsmount *mfs_mnt;
ino_t mfs_root_ino;
dev_t mfs_root_dev;
int mfs_root_usage;

#define	MFS_HASHES	64
struct mfs_map
{
	struct list_head llist;
	struct list_head glist;
	mfs_handle_t handle;
	struct dentry *l;
	struct vfsmount *mnt;
	int dcount;	/* inc: local_to_mfs ; dec: mfs_put_handle */
	int usecount;	/* inc: mfs_to_local ; dec: mfs_lput */
	unsigned long latest;
};

static struct list_head lhash[MFS_HASHES], ghash[MFS_HASHES];

static inline int
hash_dentry(struct dentry *d)
{
	return(((((unsigned int)d) >> 8) + (((unsigned int)d) >> 14) +
		(((unsigned int)d) >> 20)) % MFS_HASHES);
}

static inline int
hash_handle(mfs_handle_t l)
{
	unsigned int h = l;

	return((h+(h>>6)+(h>>12)+(h>>18)+(h>>24)) % MFS_HASHES);
}

int
mfs_conversion_init(void)
{
	unsigned int i;

	if(!mfs_unique)
	{
		mfs_root = dget(current->fs->root);
		mfs_mnt = mntget(current->fs->rootmnt);
		mfs_root_ino = mfs_root->d_inode->i_ino;
		mfs_root_dev = mfs_root->d_inode->i_dev;
		mfs_unique = ((mfs_handle_t) CURRENT_TIME) << 31;
		if(mfs_unique <= MFS_ROOT_INO)	/* 0 time? */
			mfs_unique = MFS_ROOT_INO + 1;
		for(i = 0 ; i < MFS_HASHES ; i++)
		{
			INIT_LIST_HEAD(&lhash[i]);
			INIT_LIST_HEAD(&ghash[i]);
		}
	}
	return(0);
}

void
mfs_put(struct mfs_map *m, int putuse)
{
	struct list_head *x, *nex, *prv;

	if(putuse)
		m->usecount--;
	else
		m->dcount--;
	if(m->usecount > 0 || m->dcount > 0)
		return;
	x = &m->llist;
	nex = x->next;
	prv = x->prev;
	nex->prev = prv;
	prv->next = nex;
	x = &m->glist;
	nex = x->next;
	prv = x->prev;
	nex->prev = prv;
	prv->next = nex;
	dput(m->l);
	mntput(m->mnt);
	kfree(m);
}

void
mfs_change_root(struct dentry *from, struct vfsmount *frommnt,
				struct dentry *to, struct vfsmount *tomnt)
{
	int h;
	struct list_head *head, *x;
	struct mfs_map *m, *new = NULL;
	int new_root_usage = 0;

	down(&mfsc_sem);
	if(from != mfs_root || frommnt != mfs_mnt ||
					(to == from && tomnt == frommnt))
		goto out;
	if(mfs_root_usage && !(new = (struct mfs_map *)
			kmalloc(sizeof (struct mfs_map), GFP_KERNEL)))
	{
		printk("oMFS root failed to change (ENOMEM)!\n");
		goto out;
	}

	/* is the new root listed? (if so, and no external refs, remove it) */
	head = &lhash[hash_dentry(to)];
	for(x = head->next ; x != head ; x = x->next)
	{
		m = (struct mfs_map *)((char *)x -
					offsetof(struct mfs_map, llist));
		if(m->l == to && m->mnt == tomnt)
		{
			new_root_usage = m->dcount;
			m->dcount = 1;
			mfs_put(m, 0);
			break;
		}
	}

	if(mfs_root_usage)
	{
		/* someone is actively using the old root,
		 * so we must set it up as a normal hashed entry */
		new->l = dget(from);
		new->mnt = mntget(frommnt);
		new->dcount = 0;
		new->usecount = mfs_root_usage;
		new->handle = mfs_unique++;
		new->latest = jiffies;
		h = hash_dentry(from);
		new->llist.next = lhash[h].next;
		lhash[h].next->prev = &new->llist;
		new->llist.prev = &lhash[h];
		lhash[h].next = &new->llist;
		h = hash_handle(new->handle);
		new->glist.next = ghash[h].next;
		ghash[h].next->prev = &new->glist;
		new->glist.prev = &ghash[h];
		ghash[h].next = &new->glist;
		new = NULL;
	}
	dget(to);
	mntget(tomnt);
	mfs_root = to;
	mfs_mnt = tomnt;
	mfs_root_ino = to->d_inode->i_ino;
	mfs_root_dev = to->d_inode->i_dev;
	mfs_root_usage = new_root_usage;
	dput(from);
	mntput(frommnt);
	if(new)
		kfree(new);
	out:
	up(&mfsc_sem);
}

mfs_handle_t
local_to_mfs(struct nameidata *local)
{
	struct dentry *dp = local->dentry;
	struct vfsmount *mnt = local->mnt;
	int h = hash_dentry(dp);
	struct list_head *head = &lhash[h], *x;
	struct mfs_map *m;

	down(&mfsc_sem);
	if(dp == mfs_root && mnt == mfs_mnt)
	{
		up(&mfsc_sem);
		return(MFS_PE);
	}
	for(x = head->next ; x != head ; x = x->next)
	{
		m = (struct mfs_map *)((char *)x -
					offsetof(struct mfs_map, llist));
		if(m->l == dp && mnt == mnt)
		{
			m->dcount++;
			up(&mfsc_sem);
			return(m->handle);
		}
	}
	if(!(m = (struct mfs_map *)kmalloc(sizeof (struct mfs_map),GFP_KERNEL)))
	{
		up(&mfsc_sem);
		return(0);
	}
	m->l = dget(dp);
	m->mnt = mntget(mnt);
	m->dcount = 1;
	m->usecount = 0;
	m->handle = mfs_unique++;
	m->llist.next = lhash[h].next;
	lhash[h].next->prev = &m->llist;
	m->llist.prev = &lhash[h];
	lhash[h].next = &m->llist;
	h = hash_handle(m->handle);
	m->glist.next = ghash[h].next;
	ghash[h].next->prev = &m->glist;
	m->glist.prev = &ghash[h];
	ghash[h].next = &m->glist;
	m->latest = jiffies;
	up(&mfsc_sem);
	return(m->handle);
}

int
mfs_to_local(mfs_handle_t handle, struct nameidata *local)
{
	struct list_head *head, *x, *nex, *prv;
	struct mfs_map *m;

	down(&mfsc_sem);
	if(handle == MFS_PE)
	{
		its_the_root:
		mfs_root_usage++;
		local->dentry = mfs_root;
		local->mnt = mfs_mnt;
		up(&mfsc_sem);
		return(0);
	}
	if(handle < MFS_ROOT_INO)
	{
		up(&mfsc_sem);
		return(-EINVAL);
	}
	head = &ghash[hash_handle(handle)];
	for(x = head->next ; x != head ; x = x->next)
	{
		m = (struct mfs_map *)((char *)x -
					offsetof(struct mfs_map, glist));
		if(m->handle == handle)
		{
			local->dentry = m->l;
			local->mnt = m->mnt;
			if(m->l == mfs_root && m->mnt == mfs_mnt)
				/* very rare, but possible after root change */
				goto its_the_root;
			m->usecount++;
			/* move to head of queues: */
			if(x != head->next)
			{
				nex = x->next;
				prv = x->prev;
				nex->prev = prv;
				prv->next = nex;
				x->next = head->next;
				x->prev = head;
				x->next->prev = x;
				head->next = x;
			}
			head = &lhash[hash_dentry(local->dentry)];
			x = &m->llist;
			if(x != head->next)
			{
				nex = x->next;
				prv = x->prev;
				nex->prev = prv;
				prv->next = nex;
				x->next = head->next;
				x->prev = head;
				x->next->prev = x;
				head->next = x;
			}
			m->latest = jiffies;
			up(&mfsc_sem);
			return(0);
		}
	}
	up(&mfsc_sem);
	printk("MFS: Handle %016LX not found (obsolete)\n", handle);
	return(-ESTALE);
}

void
mfs_lput(struct dentry *dp, struct vfsmount *mnt)
{
	struct list_head *head, *x;
	struct mfs_map *m;

	down(&mfsc_sem);
	if(dp == mfs_root && mnt == mfs_mnt)
	{
		mfs_root_usage--;
		up(&mfsc_sem);
		return;
	}
	head = &lhash[hash_dentry(dp)];
	for(x = head->next ; x != head ; x = x->next)
	{
		m = (struct mfs_map *)((char *)x -
					offsetof(struct mfs_map, llist));
		if(m->l == dp && m->mnt == mnt)
		{
			mfs_put(m, 1);
			up(&mfsc_sem);
			return;
		}
	}
	up(&mfsc_sem);
	printk("mfs_lput: no handle!\n");
}

int
mfs_get_handle(mfs_handle_t handle)
{
	int h = hash_handle(handle);
	struct list_head *head = &ghash[h], *x;
	struct mfs_map *m;

	if(handle < MFS_ROOT_INO)
		return(0);
	down(&mfsc_sem);
	for(x = head->next ; x != head ; x = x->next)
	{
		m = (struct mfs_map *)((char *)x -
					offsetof(struct mfs_map, glist));
		if(m->handle == handle)
		{
			m->dcount++;
			m->latest = jiffies;
			up(&mfsc_sem);
			return(0);
		}
	}
	up(&mfsc_sem);
	printk("MFS: Handle %016LX not found (obsolete) when duplicated\n",
									handle);
	return(-ESTALE);
}

int
mfs_access_handle(mfs_handle_t handle)
{
	int h = hash_handle(handle);
	struct list_head *head = &ghash[h], *x;
	struct mfs_map *m;

	if(handle < MFS_ROOT_INO)
		return(0);
	down(&mfsc_sem);
	for(x = head->next ; x != head ; x = x->next)
	{
		m = (struct mfs_map *)((char *)x -
					offsetof(struct mfs_map, glist));
		if(m->handle == handle)
		{
			m->latest = jiffies;
			up(&mfsc_sem);
			return(0);
		}
	}
	up(&mfsc_sem);
	printk("MFS: Handle %016LX not found (obsolete) when claimed\n",
									handle);
	return(-ESTALE);
}

void
mfs_put_handle(mfs_handle_t handle)
{
	int h = hash_handle(handle);
	struct list_head *head = &ghash[h], *x;
	struct mfs_map *m;

	if(handle < MFS_ROOT_INO)
		return;
	down(&mfsc_sem);
	for(x = head->next ; x != head ; x = x->next)
	{
		m = (struct mfs_map *)((char *)x -
					offsetof(struct mfs_map, glist));
		if(m->handle == handle)
		{
			mfs_put(m, 0);
			up(&mfsc_sem);
			return;
		}
	}
	up(&mfsc_sem);
	printk("MFS: Handle %016LX not found (obsolete) when released\n",
									handle);
}

int
mfs_try_to_cleanup(int force)
{
	unsigned int h;
	struct mfs_map *m;
	struct list_head *next;

	if(!mfs_unique)
		return(0);
	down(&mfsc_sem);
	if(force != 2)
	for(h = 0 ; h < MFS_HASHES ; h++)
	while(lhash[h].next != &lhash[h])
	{
		m = (struct mfs_map *)((char *)lhash[h].next -
					offsetof(struct mfs_map, llist));
		if(!force || m->usecount)
		{
			busy:
			up(&mfsc_sem);
			return(-EBUSY);
		}
	}
	for(h = 0 ; h < MFS_HASHES ; h++)
	for(next = lhash[h].next ; next != &lhash[h] ; )
	{
		m = (struct mfs_map *)((char *)next -
					offsetof(struct mfs_map, llist));
		next = next->next;
		if(m->usecount)
		{
			if(force == 2)
				m->dcount = 0;
			else
				goto busy;
		}
		else
		{
			m->dcount = 1;
			mfs_put(m, 0);
		}
	}
	up(&mfsc_sem);
	return(0);
}

void
mfs_throw_garbage(void)
{
	unsigned int h;
	struct mfs_map *m;
	struct list_head *next;
	unsigned int n = 0;

	down(&mfsc_sem);
	for(h = 0 ; h < MFS_HASHES ; h++)
	for(next = lhash[h].next ; next != &lhash[h] ; )
	{
		m = (struct mfs_map *)((char *)lhash[h].next -
					offsetof(struct mfs_map, llist));
		next = next->next;
		if(!m->usecount &&
			time_after(jiffies, m->latest + MFS_GARBAGE_TIME))
		{
			m->dcount = 1;
			mfs_put(m, 0);
			n++;
		}
	}
	up(&mfsc_sem);
	if(n)
		printk("MFS: Disposed of %d unclaimed entr%s\n", n,
							n > 1 ? "ies" : "y");
}

/* why is this not in dcache.c? */
int
is_descendant(struct dentry *dp, struct vfsmount *mnt, struct dentry *pdp,
							struct vfsmount *pmnt)
{
	spin_lock(&dcache_lock);
	while(1)
	{
		if(dp == pdp && mnt == pmnt)
		{
			spin_unlock(&dcache_lock);
			return(1);
		}
		if(dp == mnt->mnt_root)
		{
			struct vfsmount *xmnt = mnt->mnt_parent;

			if(xmnt == mnt)
				break;
			dp = mnt->mnt_mountpoint;
			mnt = xmnt;
		}
		else
			dp = dp->d_parent;
	}
	spin_unlock(&dcache_lock);
	return(0);
}

int
mfs_kill(char *filename)
{
	unsigned int h;
	struct mfs_map *m;
	struct list_head *next;
	unsigned int n = 0;
	int error;
	mm_segment_t old_fs;
	struct nameidata nd;

	old_fs = set_fs(KERNEL_DS);
	error = 0;
	if(path_init(filename, LOOKUP_POSITIVE, &nd))
		error = path_walk(filename, &nd);
	set_fs(old_fs);
	if(error)
		return(error);
	down(&mfsc_sem);
	for(h = 0 ; h < MFS_HASHES ; h++)
	for(next = lhash[h].next ; next != &lhash[h] ; )
	{
		m = (struct mfs_map *)((char *)lhash[h].next -
					offsetof(struct mfs_map, llist));
		next = next->next;
		if(is_descendant(m->l, m->mnt, nd.dentry, nd.mnt))
		{
			m->dcount = 1;
			mfs_put(m, 0);
			n++;
		}
	}
	up(&mfsc_sem);
	if(n)
		printk("MFS_KILL: Killed %d entr%s\n", n, n > 1 ? "ies" : "y");
	path_release(&nd);
	return(0);
}

/* JUST FOR DEBUGGING: */
void
mfs_print_latest(unsigned long when)
{
	int diff = (when - jiffies) / HZ;

	printk(", age=");
	if(diff < 0 || diff > 24*3600)
		printk("%ld", when);
	else if(diff == 0)
		printk("in %ld ticks", when - jiffies);
	else if(diff < 6000)
		printk("%d seconds", diff);
	else if(diff < 3600)
		printk("%d:%02d minutes", diff/60, diff % 60);
	else
		printk("%d:%02d:%02d hours", diff/3600, (diff % 3600) / 60,
								diff % 60);
}

int
dump_mfsc(void)
{
	register unsigned int h;
	struct list_head *head, *x;
	struct mfs_map *m;
	int cnt = 0;

	for(h = 0 ; h < MFS_HASHES ; h++)
	for(head = &lhash[h] , x = head->next ; x != head ; x = x->next)
	{
		cnt++;
		m = (struct mfs_map *)((char *)x -
					offsetof(struct mfs_map, llist));
		printk("%d: handle=%016LX, l=%x, ip=%x, count=%d, use=%d",
			h, m->handle,
			(int)m->l, (int)m->l->d_inode, m->dcount, m->usecount);
		if(m->l->d_inode)
			printk("\n        dev=%d/%d, ino=%ld, mode=0%o",
				MAJOR(m->l->d_inode->i_dev),
				MINOR(m->l->d_inode->i_dev),
				m->l->d_inode->i_ino, m->l->d_inode->i_mode);
		mfs_print_latest(m->latest);
		printk("\n");
	}
	return(cnt);
}

int
dump_mfsc2(void)
{
	register unsigned int h;
	struct list_head *head, *x;
	struct mfs_map *m;
	int cnt = 0;

	for(h = 0 ; h < MFS_HASHES ; h++)
	for(head = &ghash[h] , x = head->next ; x != head ; x = x->next)
	{
		cnt++;
		m = (struct mfs_map *)((char *)x -
					offsetof(struct mfs_map, glist));
		printk("%d: handle=%016LX, l=%x, ip=%x, count=%d, use=%d",
			h, m->handle,
			(int)m->l, (int)m->l->d_inode, m->dcount, m->usecount);
		if(m->l->d_inode)
			printk("\n        dev=%d/%d, ino=%ld, mode=0%o",
				MAJOR(m->l->d_inode->i_dev),
				MINOR(m->l->d_inode->i_dev),
				m->l->d_inode->i_ino, m->l->d_inode->i_mode);
		mfs_print_latest(m->latest);
		printk("\n");
	}
	return(cnt);
}
#endif /* CONFIG_MOSIX_FS */
