/* All changes since Feb 12, 2002 copyright by Moshe Bar (moshe@moshebar.com)
 * Copyright (C) 2000, 2001, Amnon Barak (amnon@cs.huji.ac.il)
 *
 * Permission to use this software is hereby granted under the terms of the
 * GNU General Public License, as published by the Free Software Foundation.
 *
 * THIS  SOFTWARE  IS  PROVIDED IN ITS  "AS IS"  CONDITION, WITH NO WARRANTY
 * WHATSOEVER. NO LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE WILL BE ACCEPTED.
 */
/*
 * Author(s): Amnon Shiloh, Moshe Bar
 */


#include <linux/sched.h>
#if defined(CONFIG_MOSIX_FS) && defined(CONFIG_MOSIX_DFSA)
#include <linux/hpcctl.h>
#include <asm/uaccess.h>
#include <linux/smp_lock.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/mfs.h>
#include <linux/mfs_socket.h>
#include <hpc/defs.h>
#include <linux/hpc.h>

#define	weight(_c,_i,_o)	((_c)*2048+(_i)+(_o))
#define	threshold		0xffe00000
/* in openMosix we never enable DEBUG_MFSCOUNT but we leave it here for documentation purposes */

#define DEBUG_MFSCOUNT
#ifdef DEBUG_MFSCOUNT
int debug_mfscount;

void do_dump_mfscount(struct task_struct *p)
{
	unsigned int i;
	struct mfs_stats *s = p->mosix.mfs_stats;

	if(!s || !s->nnodes)
	{
		printk("%s MFS Statistics for %d(%s)\n", s ? "Empty" : "No",
							p->pid, p->comm);
		return;
	}
	for(i = 0 ; i < s->nnodes ; i++)
	printk("%d: node #%d, conns=%d, inbytes=%d, outbytes=%d\n", i,
		s->nodes[i], (int)s->conns[i], (int)s->inbytes[i],
		(int)s->outbytes[i]);
}
#endif /* DEBUG_MFSCOUNT */

void
mfs_count(int w, int inbytes, int outbytes)
{
	struct task_struct *p = current;
	struct mfs_stats *s = p->mosix.mfs_stats;
	register int i;
	int64_t prev, worst = 0, best = 0;
	int victim = -1;
	long new_weight;

	if(!mfs_is_dfsa || !(p->mosix.dflags & (DREMOTE|DSTATSDOWN)))
		return;
#ifdef DEBUG_MFSCOUNT
	if(debug_mfscount > 1)
		printk("mfs_count(%d,%d,%d): ", w, inbytes, outbytes);
#endif /* DEBUG_MFSCOUNT */
	if(w == 0 || w > MOSIX_MAX)
		return;
	if(!s)
	{
		if(!(s = kmalloc(sizeof(struct mfs_stats), GFP_KERNEL)))
			return;
		p->mosix.mfs_stats = s;
		s->nnodes = 0;
	}
	new_weight = weight(1, inbytes, outbytes);
	for(i = (s->hint < s->nnodes && s->nodes[s->hint] == w) ? s->hint :
						s->nnodes - 1 ; i >= 0 ; i--)
	{
		prev = weight(s->conns[i], s->inbytes[i], s->outbytes[i]);
		if(s->nodes[i] == w)
		{
			s->conns[i]++;
			s->inbytes[i] += inbytes;
			s->outbytes[i] += outbytes;
			s->hint = i;
			if(w != PE &&
				(((((long)prev)+new_weight)^prev) & threshold))
				mosix_add_to_whereto(p, MFSBALANCE);
#ifdef DEBUG_MFSCOUNT
			if(debug_mfscount)
			{
				printk("Successfully added <%d,%d> to %d\n",
					inbytes, outbytes, w);
				if(debug_mfscount > 2)
					do_dump_mfscount(p);
			}
#endif /* DEBUG_MFSCOUNT */
			return;
		}
		if(victim == -1)
		{
			best = worst = prev;
			victim = i;
		}
		else
		{
			if(prev > best)
				best = prev;
			if(prev < worst)
			{
				worst = prev;
				victim = i;
			}
		}
	}
	if((i = s->nnodes) < MAX_MFS_STATNODES)
	{
		s->nnodes++;
		s->nodes[i] = w;
		s->conns[i] = 1;
		s->inbytes[i] = inbytes;
		s->outbytes[i] = outbytes;
		s->hint = i;
#ifdef DEBUG_MFSCOUNT
		if(debug_mfscount)
		{
			printk("New MFS stats (%d,%d) for %d\n",
				inbytes, outbytes, w);
			if(debug_mfscount > 2)
				do_dump_mfscount(p);
		}
#endif /* DEBUG_MFSCOUNT */
		return;
	}
	i = victim;
	if(--s->conns[i] <= 0 || (s->inbytes[i] -= inbytes) <= 0 ||
		(s->outbytes[i] -= outbytes) <= 0)
	{
		if(victim == --s->nnodes)
		{
#ifdef DEBUG_MFSCOUNT
			if(debug_mfscount)
			{
				printk("Anulled victim=%d\n", s->nodes[i]);
				if(debug_mfscount > 2)
					do_dump_mfscount(p);
			}
#endif /* DEBUG_MFSCOUNT */
			return;
		}
		victim = s->nnodes;
		s->nodes[i] = s->nodes[victim];
		s->conns[i] = s->conns[victim];
		s->inbytes[i] = s->inbytes[victim];
		s->outbytes[i] = s->outbytes[victim];
		if(s->hint == victim)
			s->hint = i;
#ifdef DEBUG_MFSCOUNT
		if(debug_mfscount)
		{
			printk("Anulled/replaced a victim\n");
			if(debug_mfscount > 2)
				do_dump_mfscount(p);
		}
#endif /* DEBUG_MFSCOUNT */
	}
#ifdef DEBUG_MFSCOUNT
	else if(debug_mfscount)
	{
		printk("Reduced victim=%d by <%d,%d>\n", s->nodes[i],
							inbytes, outbytes);
		if(debug_mfscount > 2)
			do_dump_mfscount(p);
	}
#endif /* DEBUG_MFSCOUNT */
	return;
}

int
mfs_add_stats(struct mfs_stats *add, struct mfs_stats *to, struct mosix_task *m)
{
	register int i;
	register unsigned int j;
	int64_t prev, worst = 0;
	int victim = -1;
#if MAX_MFS_STATNODES > BITS_PER_LONG
#error unsigned long is not enough
#else
	unsigned long bit, done = 0;
#endif /* MAX_MFS_STATNODES > BITS_PER_LONG */
	int node;
	int ret = 0;

	if(!add || !add->nnodes)
		return(0);
	if(!to)
	{
		if(!to && (!m || !(to = m->mfs_stats =
				kmalloc(sizeof(struct mfs_stats), GFP_KERNEL))))
			return(0);
		*to = *add;
		for(i = to->nnodes-1 ; i >= 0 ; i--)
		if(weight(to->conns[i], to->inbytes[i], to->outbytes[i]) &
								threshold)
			return(1);
		return(0);
	}
	for(i = 0 ; i < to->nnodes ; i++)
	{
		node = to->nodes[i];
		for(j = 0 ; j < add->nnodes ; j++)
		if(add->nodes[j] == node)
		{
			prev = weight(to->conns[i], to->inbytes[i],
							to->outbytes[i]);
			to->conns[i] += add->conns[i];
			to->inbytes[i] += add->inbytes[i];
			to->outbytes[i] += add->outbytes[i];
			if(!ret && node != PE &&
				(weight(to->conns[i], to->inbytes[i],
					to->outbytes[i]) ^ prev) & threshold)
				ret = 1;
			done |= (1 << (j-1));
			break;
		}
	}
	for(bit = 1 , j = 0 ; j < add->nnodes ; j++ , bit <<= 1)
	if(!(done & bit))
	{
		if(to->nnodes < MAX_MFS_STATNODES)
		{
			i = to->nnodes++;
			to->nodes[i] = add->nodes[j];
			to->conns[i] = add->conns[j];
			to->inbytes[i] = add->inbytes[j];
			to->outbytes[i] = add->outbytes[j];
			if(!ret && (weight(to->conns[i], to->inbytes[i],
				to->outbytes[i]) & threshold))
				ret = 1;
			continue;
		}
		for(i = 0 ; i < MAX_MFS_STATNODES ; i++)
		{
			prev = weight(add->conns[i], add->inbytes[i],
							add->outbytes[i]);
			if(i == 0 || prev < worst)
			{
				victim = i;
				worst = prev;
			}
		}
		if(!ret && (weight(add->conns[i], add->inbytes[i],
						add->outbytes[i]) & threshold))
			ret = 1;
		to->nodes[victim] = add->nodes[i];
		to->conns[victim] = add->conns[i];
		to->inbytes[victim] = add->inbytes[i];
		to->outbytes[victim] = add->outbytes[i];
	}
	return(ret);
}

#ifdef DEBUG_MFSCOUNT
void
dump_mfscount(int pid)
{
	struct task_struct *p;

	if(!pid)
		p = current;
	else
		p = find_any_task_by_pid(pid);
	if(p)
		do_dump_mfscount(p);
	else
		printk("No Such Task (%d)\n", pid);
}
#endif /* DEBUG_MFSCOUNT */

#endif /* CONFIG_MOSIX_FS && CONFIG_MOSIX_DFSA */
