/*
 * Copyright (C) 2000, 2001, Amnon Barak (amnon@cs.huji.ac.il)
 *
 * Permission to use this software is hereby granted under the terms of the
 * GNU General Public License, as published by the Free Software Foundation.
 *
 * THIS  SOFTWARE  IS  PROVIDED IN ITS  "AS IS"  CONDITION, WITH NO WARRANTY
 * WHATSOEVER. NO LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE WILL BE ACCEPTED.
 */
/*
 * Author(s): Amnon Shiloh.
 */

#include <linux/sched.h>
#ifdef CONFIG_MOSIX_FS
#include <asm/uaccess.h>
#include <linux/smp_lock.h>
#include <linux/slab.h>
#include <linux/mfs.h>
#include <linux/mfs_socket.h>
#include <hpc/routines.h>

#define	MFS_MINSERVERS	2
#define	MFS_MAXDEDICATED 6
#define	MFS_MAXSERVERS	14
#define	MFS_CONTACT_KEEPALIVE (60*HZ)
#define	MFS_MAXQ	100	/* do not increase - leave space under 128 */

spinlock_t mfs_servq_lock = SPIN_LOCK_UNLOCKED;

struct task_struct *mfs_main_server_task;
int n_mfs_servers;
int mfs_total_services;

struct mfs_server_contact
{
	struct mfs_server_contact *next;
	struct socket *sock;
	int caller;
	int serial;
	unsigned long last_use;
};

/* service queue: */
struct mfs_server_contact *queued_services;
struct mfs_server_contact *last_queued_service;
int mfs_qlen;

/* past connections that are still kept open: */
struct mfs_server_contact *mfs_idle_connections;

/* servers waiting: */
DECLARE_WAIT_QUEUE_HEAD(wait_to_serve_mfs);
atomic_t waiting_to_serve_mfs = ATOMIC_INIT(0);
int mfs_idle_servers;

/* main server waiting: */
DECLARE_WAIT_QUEUE_HEAD(wait_for_mfs_calls);
volatile int waiting_for_mfs_calls;
volatile int mfs_has_new_calls;

int
mfs_server_send(struct mfs_server_contact *s, int type, void *header, int hdln,
			void *dat1, int len1, void *dat2, int len2)
{
	return(mfs_send(s->sock, 0, s->serial, type,
					header, hdln, dat1, len1, dat2, len2));
}

int
mfs_server_receive(struct mfs_server_contact *s,
					void **header, void **dat1, void **dat2)
{
	return(mfs_receive(s->sock, MFS_MUST_MATCH, &s->serial, NULL,
							header, dat1, dat2));
}

void
mfs_serve1(struct mfs_server_contact *s)
{
	int type, more, changed;
	void *header;
	void *dat1;
	void *dat2;
	static int n_dedicated;

	mfs_set_main_ownership(s->sock, 0);
	again:
	type = mfs_receive(s->sock, MFS_FIRST_RECEIVE, &s->serial, NULL,
							&header, &dat1, &dat2);
	if(type == MFS_CLOSE || type < MFS_NONE)
	{
		mfs_close_socket(s->sock);
if(mfs_debug > 1)
printk("mfs_serve1: closing socket %x by client request!\n", (int)s->sock);
		s->sock = NULL;
	}
	else
		mfs_serve(s, type, header, dat1, dat2);
	if(header)
		kfree(header);
	if(!s->sock)
		return;
	spin_lock(&mfs_servq_lock);
	more = 1 + (n_dedicated < MFS_MAXDEDICATED);
	if(more == 2)
		n_dedicated++;
	spin_unlock(&mfs_servq_lock);
	changed = mfs_set_main_ownership(s->sock, more);
	if(more == 2)
	{
		spin_lock(&mfs_servq_lock);
		n_dedicated--;
		spin_unlock(&mfs_servq_lock);
	}
	if(!changed)
	{
		if(mfs_debug)
			printk("mfs_serv1: serving in chain!\n");
		goto again;
	}
}

/* trim_idle_connections must be called with mfs_servq_lock! */
void
trim_idle_connections(void)
{
#define	HOGS_CONSIDERED	10
	struct mfs_server_contact *s, *prev, *next, *dispose = NULL;
	unsigned int nhogs = 0;
	int hogpe[HOGS_CONSIDERED];
	int hogs[HOGS_CONSIDERED];
	unsigned int i, least;
	int caller, leastval, penalty, bar;
	unsigned int disposed = 0;

	/* compile the hog-table */
	for(prev = NULL, s = mfs_idle_connections ; s ; prev = s , s = s->next)
	{
		caller = s->caller;
		for(i = 0 ; i < nhogs ; i++)
		if(hogpe[i] == caller)
		{
			hogs[i]++;
			goto next_connection;
		}
		if(i < HOGS_CONSIDERED)
		{
			hogpe[i] = caller;
			hogs[i] = 1;
			nhogs++;
			goto next_connection;
		}
		leastval = 10000000;
		least = 0;
		for(i = 0 ; i < HOGS_CONSIDERED ; i++)
		if(hogs[i] < leastval)
		{
			least = i;
			leastval = hogs[i];
		}
		hogpe[least] = caller;
		hogs[least] = 1;
		next_connection:;
	}
	/* calculate penalties, store in the "serial" field (unused for idle) */
	/* penalty increases for old sockets and for hogs */
	for(s = mfs_idle_connections ; s ; s = s->next)
	{
		penalty = abs((long)jiffies - (long)(s->last_use));
		if(penalty > 2*MFS_CONTACT_KEEPALIVE)	/* avoid overflow */
			penalty = 2*MFS_CONTACT_KEEPALIVE;
		penalty = mfs_total_services * penalty / MFS_CONTACT_KEEPALIVE; 
		caller = s->caller;
		for(i = 0 ; i < nhogs ; i++)
		if(hogpe[i] == caller)
		{
			penalty *= hogs[i];
			break;
		}
		s->serial = penalty;
	}
	/* collect old sockets, more aggressively as services are exhausted */
	for(bar = MFS_MAXQ/2 ;
		mfs_idle_connections && mfs_total_services - disposed > MFS_MAXQ
							; bar = bar * 2 / 3)
	for(prev = NULL, s = mfs_idle_connections ; s ; )
	if(s->serial >= bar)
	{
		next = s->next;
		if(prev)
			prev->next = next;
		else
			mfs_idle_connections = next;
if(mfs_debug)
printk("trim_idle_connections: penalty=%d, closing socket %x\n", s->serial,
(int)s->sock);
		s->next = dispose;
		dispose = s;
		disposed++;
		s = next;
	}
	else
	{
		prev = s;
		s = s->next;
	}
	/* dispose of those sockets */
	mfs_total_services -= disposed;
	spin_unlock(&mfs_servq_lock);
	for(; dispose ; dispose = next)
	{
		next = dispose->next;
		mfs_close_socket(dispose->sock);
		kfree(dispose);
	}
	spin_lock(&mfs_servq_lock);
}

void
mfs_recycle_service(struct mfs_server_contact *s)
{
	spin_lock(&mfs_servq_lock);
	if(!s->sock)
	{
		mfs_total_services--;
		spin_unlock(&mfs_servq_lock);
		kfree(s);
		return;
	}
	if(mfs_total_services >= MFS_MAXQ)
		trim_idle_connections();
	if(mfs_total_services < MFS_MAXQ)
	{
		s->last_use = jiffies;
		s->next = mfs_idle_connections;
		mfs_idle_connections = s;
		spin_unlock(&mfs_servq_lock);
		return;
	}
if(mfs_debug)
printk("mfs_recycle_service: closing socket %x (too many connections already)\n", (int)s->sock);
	mfs_total_services--;
	spin_unlock(&mfs_servq_lock);
	mfs_close_socket(s->sock);
	kfree(s);
}

int
mfs_server(void *nothing)
{
	struct mfs_server_contact *s;
	struct task_struct *p = current;
	extern struct task_struct *child_reaper;

	strcpy(p->comm, "mfs_server");
	p->exit_signal = SIGCHLD;
	p->p_opptr = child_reaper;
	write_lock_irq(&tasklist_lock);
	REMOVE_LINKS(p);
	p->p_pptr = p->p_opptr;
	SET_LINKS(p);
	p->fs->umask = 0;
	write_unlock_irq(&tasklist_lock);
	set_fs(KERNEL_DS);	/* all server I/O always to kernel */
	spin_lock(&mfs_servq_lock);
	mfs_idle_servers--;
	while(1)
	{
		while((s = queued_services))
		{
			mfs_qlen--;
			if(!(queued_services = s->next))
				last_queued_service = NULL;
			spin_unlock(&mfs_servq_lock);
			mfs_serve1(s);
			mfs_recycle_service(s);
			spin_lock(&mfs_servq_lock);
		}
		if(n_mfs_servers > MFS_MINSERVERS || MFS_PE == MFS_UNKNOWN_PE)
		{
			n_mfs_servers--;
			spin_unlock(&mfs_servq_lock);
			return(0);		/* and exit */
		}
		mfs_idle_servers++;
		atomic_inc(&waiting_to_serve_mfs);
		spin_unlock(&mfs_servq_lock);
		interruptible_sleep_on(&wait_to_serve_mfs);
		if(p->sigpending)
		{
			spin_lock_irq(&p->sigmask_lock);
			flush_signals(p);
			spin_unlock_irq(&p->sigmask_lock);
		}
		spin_lock(&mfs_servq_lock);
		atomic_dec(&waiting_to_serve_mfs);
		mfs_idle_servers--;
	}
}

void
mfs_new_request_arrived(void)
{
if(mfs_debug > 1)
printk("mfs_new_request_arrived!\n");
	mfs_has_new_calls = 1;
	if(waiting_for_mfs_calls)
		wake_up_interruptible(&wait_for_mfs_calls);
}

/* mfs_add_to_queue must be called with mfs_servq_lock! */
static inline void
mfs_add_to_queue(struct mfs_server_contact *s)
{
	s->next = NULL;
	mfs_qlen++;
	if(last_queued_service)
		last_queued_service->next = s;
	else
		queued_services = s;
	last_queued_service = s;
	if(atomic_read(&waiting_to_serve_mfs))
		wake_up_interruptible(&wait_to_serve_mfs);
}

int
mfs_main_server(void *nothing)
{
	register struct task_struct *p = current;
	struct socket *mainsock = NULL, *sock;
	struct mfs_server_contact *s, *prev, *next, *dispose = NULL;
	int caller;
	unsigned long last_trim = 0;
	int client_pid = 0;
	int success;
	DECLARE_WAITQUEUE(wait, p);

	mfs_main_server_task = p;
	common_daemon_setup("oMfs_main_server", 1);
	while(client_pid <= 0 || !mainsock)
	{
		if(client_pid <= 0)
			client_pid = kernel_thread(mfs_client_daemon, NULL, 0);
		if(!mainsock)
			mainsock = mfs_open_main_socket();
		if(client_pid > 0 && mainsock)
			break;
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(MFS_CONTACT_KEEPALIVE / 3);
		spin_lock_irq(&p->sigmask_lock);
		flush_signals(p);
		spin_unlock_irq(&p->sigmask_lock);
	}
	mainloop:
	spin_lock(&mfs_servq_lock);
	if(MFS_PE != MFS_UNKNOWN_PE)
	while(n_mfs_servers < MFS_MINSERVERS)
	{
		n_mfs_servers++;
		mfs_idle_servers++;
		spin_unlock(&mfs_servq_lock);
		success = (kernel_thread(mfs_server, NULL, 0) >= 0);
		spin_lock(&mfs_servq_lock);
		if(!success)
		{
			n_mfs_servers--;
			mfs_idle_servers--;
			break;
		}
	}
	if(mfs_has_new_calls)
	{
		spin_unlock(&mfs_servq_lock);
		mfs_has_new_calls = 0;
		while((sock = mfs_new_connection(mainsock, &caller)))
		if(MFS_PE == MFS_UNKNOWN_PE)
			mfs_close_socket(sock);
		else
		{
if(mfs_debug)
printk("mfs_main_server: new connection=%x (mainsock=%x)\n", (int)sock, (int)mainsock);
			if((s = (struct mfs_server_contact *)kmalloc(
				sizeof(struct mfs_server_contact), GFP_KERNEL)))
			{
				spin_lock(&mfs_servq_lock);
				mfs_total_services++;
				s->sock = sock;
				s->caller = caller;
				mfs_add_to_queue(s);
				spin_unlock(&mfs_servq_lock);
			}
			else
				mfs_close_socket(sock);
		}
		spin_lock(&mfs_servq_lock);
		for(prev = NULL, s = mfs_idle_connections ; s ; )
		if(MFS_PE == MFS_UNKNOWN_PE || mfs_socket_has_data(s->sock))
		{
			next = s->next;
			if(prev)
				prev->next = next;
			else
				mfs_idle_connections = next;
			if(MFS_PE == MFS_UNKNOWN_PE)
			{
				s->next = dispose;
				dispose = s;
				mfs_total_services--;
			}
			else
			{
if(mfs_debug)
printk("mfs_main_server: new request on old sock (%x)\n", (int)s->sock);
				mfs_add_to_queue(s);
			}
			s = next;
		}
		else
		{
			prev = s;
			s = s->next;
		}
		while(mfs_qlen > mfs_idle_servers &&
			n_mfs_servers < MFS_MAXSERVERS &&
			MFS_PE != MFS_UNKNOWN_PE)
		{
if(mfs_debug > 1)printk("mfs_main_server adding a server\n");
			n_mfs_servers++;
			mfs_idle_servers++;
			spin_unlock(&mfs_servq_lock);
			success = (kernel_thread(mfs_server, NULL, 0) >= 0);
			spin_lock(&mfs_servq_lock);
			if(!success)
			{
				n_mfs_servers--;
				mfs_idle_servers--;
				if(!n_mfs_servers)
				{
					/* too bad - just shut all down */
					while((s = queued_services))
					{
						queued_services = s->next;
						mfs_qlen--;
						mfs_total_services--;
						spin_unlock(&mfs_servq_lock);
						mfs_close_socket(s->sock);
						kfree(s);
						spin_lock(&mfs_servq_lock);
					}
				}
				break;
			}
		}
	}
	if(abs((long)jiffies - (long)last_trim) > HZ)
	{
		for(prev = NULL, s = mfs_idle_connections ; s ; )
		if(abs((long)jiffies - (long)(s->last_use)) >=
							MFS_CONTACT_KEEPALIVE)
		{
			next = s->next;
			if(prev)
				prev->next = next;
			else
				mfs_idle_connections = next;
			if(mfs_socket_has_data(s->sock))
				mfs_add_to_queue(s);
			else
			{
if(mfs_debug > 1)
printk("mfs_main_server: closing old socket %x\n", (int)s->sock);
				s->next = dispose;
				dispose = s;
				mfs_total_services--;
			}
			s = next;
		}
		else
		{
			prev = s;
			s = s->next;
		}
	}
	spin_unlock(&mfs_servq_lock);
	last_trim = jiffies;
	for(; dispose ; dispose = next)
	{
		next = dispose->next;
		mfs_close_socket(dispose->sock);
		kfree(dispose);
	}
	if(!mfs_has_new_calls)
	{
		add_wait_queue(&wait_for_mfs_calls, &wait);
		set_current_state(TASK_INTERRUPTIBLE);
		waiting_for_mfs_calls = 1;
		if(!mfs_has_new_calls)
			schedule_timeout(MFS_CONTACT_KEEPALIVE / 3);
		waiting_for_mfs_calls = 0;
		remove_wait_queue(&wait_for_mfs_calls, &wait);
		set_current_state(TASK_RUNNING);
	}
	spin_lock_irq(&p->sigmask_lock);
	flush_signals(p);
	spin_unlock_irq(&p->sigmask_lock);
	mfs_throw_garbage();
	goto mainloop;
}

void
mfs_make_servers(void)
{
	kernel_thread(mfs_main_server, NULL, 0);
}

#endif /* CONFIG_MOSIX_FS */
