/* Changes since Feb 12, 2002 by Moshe Bar
 * Copyright (C) 2000, 2001, Amnon Barak (amnon@cs.huji.ac.il)
 *
 * Permission to use this software is hereby granted under the terms of the
 * GNU General Public License, as published by the Free Software Foundation.
 *
 * Author(s): Amnon Shiloh, Moshe Bar
 * used code by Oren Laadan.
 */

#include <linux/sched.h>
#ifdef CONFIG_MOSIX_FS
#include <net/tcp.h>
#include <linux/hpc.h>
#include <linux/slab.h>
#include <linux/mfs.h>
#include <linux/mfs_socket.h>
#include <linux/poll.h>
#include <hpc/routines.h>
#include <hpc/protocol.h>

/* meanwhile, we assume openMosix uses only IPv4 addresses */
/* when openMosix can use IPv6, this file must also be upgraded */

#ifdef CONFIG_MOSIX_UDB
#define	MFS_CONNECT_TIMO	(100*HZ)
#define	MFS_RECEIVE_TIMO	(100*HZ)
#else
#define	MFS_CONNECT_TIMO	(10*HZ)
#define	MFS_RECEIVE_TIMO	(10*HZ)
#endif /* CONFIG_MOSIX_UDB */

int
mfs_common_socket_setup(struct socket *sock)
{
	mm_segment_t oldfs = get_fs();
	int error = 0;
	int val;

	set_fs(KERNEL_DS);
	val = 32768+1024;
	sock_setsockopt(sock, SOL_SOCKET, SO_SNDBUF, (char *)&val, sizeof(val));
	sock_setsockopt(sock, SOL_SOCKET, SO_RCVBUF, (char *)&val, sizeof(val));
	val = MOSIX_CONNECTION_KEEPALIVE_INTERVAL;
	sock->ops->setsockopt(sock, IPPROTO_TCP, TCP_KEEPINTVL,
						(char *) &val, sizeof(val));
	val = MOSIX_CONNECTION_KEEPALIVE_MAXTRIES;
	sock->ops->setsockopt(sock, IPPROTO_TCP, TCP_KEEPCNT,
						(char *) &val, sizeof(val));
	val = MOSIX_CONNECTION_KEEPALIVE_TOTAL;
	sock->ops->setsockopt(sock, IPPROTO_TCP, TCP_KEEPIDLE,
						(char *) &val, sizeof(val));
	val = 1;
	sock->ops->setsockopt(sock, IPPROTO_TCP, TCP_NODELAY,
						(char *)&val, sizeof(val));
	error = sock_setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE,
						(char *)&val, sizeof(val));
	set_fs(oldfs);
	set_bit(SOCK_INTER_MOSIX, &sock->flags);
	return(error);
}

struct socket *
mfs_open_main_socket(void)
{
	struct socket *sock;
	struct sockaddr_in addr;

	deeper_sleep();

	memset(&addr, 0, sizeof(struct sockaddr_in));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = INADDR_ANY;
	addr.sin_port = MFS_MAIN_PORT;
	if(sock_create(AF_INET, SOCK_STREAM, IPPROTO_TCP, &sock))
		goto failed;

	if(mfs_common_socket_setup(sock))
	{
		goto failed;
	}

	if(sock->ops->bind(sock, (struct sockaddr *)&addr, sizeof(addr)))
	{
		goto failed;
	}
	if(sock->ops->listen(sock, SOMAXCONN))
	{
		goto failed;
	}

	lighter_sleep();
	spin_lock_irq(&skown_lock);
	sock->sk->owner = mfs_main_server_task;
	spin_unlock_irq(&skown_lock);
	return(sock);

failed:
	lighter_sleep();
	if (sock)
		sock_release(sock);
	return(NULL);
}

struct socket *
mfs_new_connection(struct socket *parent, int *caller)
{
	struct socket *sock;
	struct mosix_addr mosaddr;
	int val = sizeof(mosaddr.saddr);

	if(!(parent->ops->poll(NULL, parent, NULL) & POLLIN))
		return(NULL);
	if (!(sock = sock_alloc()))
		return(NULL);
	sock->type = parent->type;
	sock->ops = parent->ops;
	if(sock->ops->accept(parent, sock, O_NONBLOCK))
		goto failed;
	if (sock->ops->getname(sock, &mosaddr.saddr, &val, 1))
		goto failed;
	if(!(*caller = net_to_mos(&mosaddr)))
	{
		comm_report_violation("MFS", &mosaddr.saddr);
		goto failed;
	}
	if(mfs_common_socket_setup(sock))
		goto failed;
	spin_lock_irq(&skown_lock);
	sock->sk->owner = mfs_main_server_task;
	spin_unlock_irq(&skown_lock);
	return(sock);
	failed:
	sock_release(sock);
	return(NULL);
}

struct socket *
mfs_connect_to(int to)
{
	struct task_struct *p = current;
	int error;
	struct socket *sock = NULL;
	struct sockaddr addr;
	unsigned long timo = MFS_CONNECT_TIMO;
	DECLARE_WAITQUEUE(wait, p);

	deeper_sleep();
	if(!mos_to_net(to, &addr))
	{
		error = -EREMOTE;
		goto failed;
	}
	((struct sockaddr_in *)&addr)->sin_port = MFS_MAIN_PORT;
	if((error = sock_create(AF_INET, SOCK_STREAM, IPPROTO_TCP, &sock)))
		goto failed;
	if((error = mfs_common_socket_setup(sock)))
		goto failed;
	sock->ops->connect(sock, &addr, sizeof(addr), O_NONBLOCK);
	add_wait_queue(sock->sk->sleep, &wait);
	while (sock->state != SS_CONNECTED)
	{
		set_current_state(TASK_INTERRUPTIBLE);
		error = sock->ops->connect(sock, &addr, sizeof(addr),
								O_NONBLOCK);
		if (error != -EALREADY || (error = sock_error(sock->sk)))
			break;
		timo = schedule_timeout(timo);
		if (timo <= 0)
		{
			error = -EAGAIN;
			break;
		}
	}
	remove_wait_queue(sock->sk->sleep, &wait);
	set_current_state(TASK_RUNNING);

	if (error)
		goto failed;
	if (sock->sk->err)
	{
		error = sock_error(sock->sk);	/* cleans error.. */
		goto failed;
	}
	lighter_sleep();
	return (sock);

failed:
	lighter_sleep();
	if (sock)
		sock_release(sock);
	return (ERR_PTR(error));
}

inline int
mfs_socket_has_data(struct socket *sock)
{
	return((sock->ops->poll(NULL, sock, 0) &
				(POLLRDNORM|POLLIN|POLLHUP|POLLERR)) != 0);
}

/*
 * "mfs_socket_is_closed" is only called by the client on inactive sockets:
 * we never expect server-initiated data, so POLLIN implies a closed socket.
 */
int
mfs_socket_is_closed(struct socket *sock)
{
if(mfs_debug > 1)printk("socket_is_closed(%x): poll=0x%x\n", (int)sock, sock->ops->poll(NULL, sock, NULL));
	return((sock->ops->poll(NULL, sock, NULL) &
						(POLLHUP|POLLERR|POLLIN)) != 0);
}

int
mfs_send(struct socket *sock, int first, int serial, int type,
	void *header, int hdsz, void *dat1, int dat1sz, void *dat2, int dat2sz)
{
	struct mfs_message_header hd;
	struct iovec iov[5];
	struct msghdr msg = { NULL, 0, iov, 0, NULL, 0, MSG_NOSIGNAL };
	mm_segment_t oldfs;
	unsigned int niov = 1;
	int size = 0;
	int error = -ENOMEM;

	hd.serial = serial;
	hd.first = first;
	hd.version = MFS_VERSION;
	hd.caller = PE;
	hd.type = type;
	hd.hdsz = hdsz;
	hd.dat1sz = dat1sz;
	hd.dat2sz = dat2sz;
	iov[0].iov_base = &hd;
	iov[0].iov_len = sizeof(hd);
	if(hdsz)
	{
		iov[1].iov_base = header;
		iov[1].iov_len = hdsz;
		niov++;
	}
	if(dat1sz)
	{
		iov[niov].iov_base = dat1;
		iov[niov++].iov_len = dat1sz;
	}
	if(dat2sz)
	{
		iov[niov].iov_base = dat2;
		iov[niov++].iov_len = dat2sz;
	}
	size = sizeof(hd) + hdsz + dat1sz + dat2sz;
	msg.msg_iovlen = niov;

	deeper_sleep();
	oldfs = get_fs();
	set_fs(KERNEL_DS);
	error = sock->ops->sendmsg(sock, &msg, size, 0);
	set_fs(oldfs);
	lighter_sleep();

	return (error < 0 ? error : 0);
}

int
mfs_do_receive(struct socket *sock, void *base, int len)
{
	struct task_struct *p = current;
	int n = 0;
	struct iovec iov;
	struct msghdr msg = { NULL, 0, &iov, 1, NULL, 0,
				MSG_NOSIGNAL | MSG_DONTWAIT};
	int done = 0;
	unsigned long timo = MFS_RECEIVE_TIMO;
	DECLARE_WAITQUEUE(wait, p);

	add_wait_queue(sock->sk->sleep, &wait);
	while(done < len)
	{
		set_current_state(TASK_INTERRUPTIBLE);
		p->mosix.ignoreoldsigs = 1;
		p->sigpending = 0;
		if(!mfs_socket_has_data(sock))
		{
			if(timo <= 0)
			{
				len = -EPIPE;
				break;
			}
			timo = schedule_timeout(timo);
			continue;
		}
		iov.iov_base = (void *)((char *)base + done);
		iov.iov_len = len - done;
		msg.msg_iov = &iov;
		msg.msg_iovlen = 1;
		n = sock->ops->recvmsg(sock, &msg, len-done, msg.msg_flags, 0);
		if (n == -EAGAIN || n == -ERESTARTSYS || n == -EINTR)
			continue;
		if(n <= 0)
		{
			len = n < 0 ? n : ECONNRESET;
			break;
		}
		done += n;
	}
	remove_wait_queue(sock->sk->sleep, &wait);
	set_current_state(TASK_RUNNING);
	recalc_sigpending(p);
	return (len);
}

int
mfs_receive(struct socket *sock, int policy, int *serial,
		struct mfs_message_header *preread_header,
					void **header, void **dat1, void **dat2)
{
	struct mfs_message_header hd;
	void *contents = NULL;
	int tlen;
	mm_segment_t oldfs;
	int error;
	unsigned int need_another;

	deeper_sleep();
	oldfs = get_fs();
	set_fs(KERNEL_DS);

	next_message:
	need_another = 0;

	if(preread_header && policy != MFS_JUST_HEADER)
	{
		hd = *preread_header;
		preread_header = NULL;
		error = 0;
	}
	else if ((error = mfs_do_receive(sock, &hd, sizeof(hd))) != sizeof(hd))
	{
if(mfs_debug)
printk("%s-mfs_receive: asked for header(%d), got %d\n", desc_mostask(NULL), sizeof(hd), error);
		if (error >= 0)
			error = -EPIPE;
		goto failed;
	}
if(mfs_debug)
printk("%s-mfs_do_receive: asked and got header: ", desc_mostask(NULL));
if(mfs_debug)
printk("version=%d, first=%d, caller=%d, serial=%d, type=%d, size=%d/%d/%d\n", hd.version,hd.first, hd.caller, hd.serial, hd.type, hd.hdsz, hd.dat1sz, hd.dat2sz);
	if(hd.version != MFS_VERSION)
	{
		static int limit;

		if(limit++ < 5)
		{
			printk("MFS received request from wrong version:\n");
			printk("Mine is %d, theirs (%d) is %d\n", MFS_VERSION,
							hd.caller, hd.version);
			error = -EINVAL;
			goto failed;
		}
	}
	error = 0;
	switch(policy)
	{
		case MFS_JUST_HEADER:
			if(hd.serial != *serial)
			{
				need_another = 1;
				break;
			}
			*preread_header = hd;
			goto done;
		case MFS_SKIP_OTHERS:
			if(hd.serial != *serial)
				need_another = 1;
			break;
		case MFS_MUST_MATCH:
			if(hd.first || hd.serial != *serial)
			{
				error = -EPIPE;
				goto failed;
			}
			break;
		case MFS_FIRST_RECEIVE:
			if(hd.first)
				*serial = hd.serial;
			else
			{
				error = -EPIPE;
				goto failed;
			}
			break;
	}
	tlen = hd.hdsz + hd.dat1sz + hd.dat2sz;

	if (tlen)
	{
		if(!(contents = kmalloc(tlen, GFP_KERNEL)))
		{
			error = -ENOMEM;
			goto failed;
		}
		if ((error = mfs_do_receive(sock, contents, tlen)) != tlen) {
if(mfs_debug)
printk("%s-mfs_receive: asked for body(%d), got %d\n", desc_mostask(NULL), tlen, error);
			if (error >= 0)
				error = -EFAULT;
			goto failed;
		}
if(mfs_debug>1)
printk("%s-mfs_receive: asked and got body(%d)\n", desc_mostask(NULL), tlen);
	}
	error = 0;

done:
failed:
	lighter_sleep();
	set_fs(oldfs);
	if(error)
	{
		if(contents)
			kfree(contents);
		if(header)
			*header = NULL;
		if(dat1)
			*dat1 = NULL;
		if(dat2)
			*dat2 = NULL;
if(mfs_debug)
printk("%s-mfs_receive: error=%d\n", desc_mostask(NULL), error);
		return(error);
	}
	else if(need_another)
	{
		need_another = 0;
if(mfs_debug)
printk("%s-mfs_receive: need another!\n", desc_mostask(NULL));
		if(contents)
		{
			kfree(contents);
			contents = NULL;
		}
		goto next_message;
	}
	if(header)
		*header = contents;
	if(dat1)
		*dat1 = (void *)(((char *)contents) + hd.hdsz);
	if(dat2)
		*dat2 = (void *)(((char *)contents) + hd.hdsz + hd.dat1sz);
	return(hd.type);
}

int
mfs_set_main_ownership(struct socket *sock, int main)
{
	int timo = HZ;

	spin_lock_irq(&skown_lock);
	if(!main)
	{
		sock->sk->owner = NULL;
		spin_unlock_irq(&skown_lock);
		return(1);
	}
	sock->sk->owner = current;
	set_bit(SOCK_WAIT_MFSDATA, &sock->flags);
	spin_unlock_irq(&skown_lock);
	while(1)
	{
		current->state = TASK_INTERRUPTIBLE;
		if(sock->ops->poll(NULL, sock, 0) & POLLIN)
		{
			spin_lock_irq(&skown_lock);
			clear_bit(SOCK_WAIT_MFSDATA, &sock->flags);
			sock->sk->owner = NULL;
			spin_unlock_irq(&skown_lock);
			current->state = TASK_RUNNING;
			return(0);
		}
		if(main == 1 || timo <= 0)
		{
			spin_lock_irq(&skown_lock);
			clear_bit(SOCK_WAIT_MFSDATA, &sock->flags);
			sock->sk->owner = NULL;
			current->state = TASK_RUNNING;
			spin_unlock_irq(&skown_lock);
			sock->sk->owner = mfs_main_server_task;
			mfs_new_request_arrived();
			return(1);
		}
		timo = schedule_timeout(timo);
	}
}

void
mfs_close_socket(struct socket *sock)
{
	spin_lock_irq(&skown_lock);
	if(sock->sk)
		sock->sk->owner = NULL;
	clear_bit(SOCK_INTER_MOSIX, &sock->flags);
	spin_unlock_irq(&skown_lock);
	if(sock->ops)
		sock->ops->shutdown(sock, 2);
	sock_release(sock);
}

#endif /* CONFIG_MOSIX_FS */
